# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import sys
import b
import c
import os
import a


class Tree:
    branches: int
    height_cubits: float


def do_something_really_interesting_with_a_lot_of_arguments(
    arg_foo, arg_a: str, arg_b: int = 0, **kwargs
) -> None:
    call_some_func()
    print("a str")



#!/usr/bin/env python
# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
Zephyr CLI

The 'zephyr' subcommand for the Pigweed CLI provides functionality for
integrating a Pigweed and Zephyr application.
"""

import argparse
import os
import re
import subprocess
from pathlib import Path
import logging
import yaml

from pw_env_setup_zephyr.argparser import add_parser_arguments


_LOG = logging.getLogger('pw_zephyr.zephyr')


def get_parser() -> argparse.ArgumentParser:
    """Get a parser associated with this command."""
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser = add_parser_arguments(parser)
    return parser


def maybe_append_remote(remotes: list, name: str, path: Path) -> bool:
    """Try appending a remote entry if it exists and is valid."""
    if not path.exists():
        _LOG.debug("'%s' does not exist, can't add '%s'", str(path), name)
        return False
    if not path.is_dir():
        _LOG.debug("'%s' is not a directory, can't add '%s'", str(path), name)
        return False

    result = subprocess.run(
        ['git', 'remote', '-v'],
        stdout=subprocess.PIPE,
        cwd=path,
    )
    if result.returncode != 0:
        _LOG.error("Failed to find remotes for '%s' in '%s'", name, str(path))
        return False
    output = result.stdout.decode('utf-8')
    _LOG.debug("git remotes=\n%s", output)
    url_match = re.match(
        r'^\w+\s+([\w@\:\/\-\.]+)\s+\(fetch\).*$', output, re.MULTILINE
    )
    if not url_match:
        _LOG.error(
            "Failed to find (fetch) remote for '%s' in '%s'", name, str(path)
        )
        return False
    remotes.append(
        {
            'name': name,
            'url-base': url_match.group(1),
        }
    )
    return True


def append_project(projects: list, name: str, path: Path) -> None:
    """Append a project entry to the list"""
    result = subprocess.run(
        ['git', 'log', '--pretty=format:"%H"', '-n', '1'],
        stdout=subprocess.PIPE,
        cwd=path,
    )
    if result.returncode != 0:
        _LOG.error(
            "Failed to find hash branch for '%s' in '%s'", name, str(path)
        )
        return
    output: str = result.stdout.decode('utf-8').strip().strip('"')
    _LOG.debug("git current hash is '%s'", output)
    if output == '':
        _LOG.error("'%s' has no commit hash", str(path))
        return

    projects.append(
        {
            'name': name,
            'remote': name,
            'revision': output,
            'path': str(
                path.relative_to(os.path.commonpath([path, os.getcwd()]))
            ),
            'import': name == 'zephyr',
        }
    )


def generate_manifest() -> dict:
    """Generate a dictionary that matches a West manifest yaml format."""
    pw_root_dir = Path(os.environ['PW_ROOT'])
    zephyr_root_dir = pw_root_dir / 'environment' / 'packages' / 'zephyr'
    remotes: list = []
    projects: list = []
    if maybe_append_remote(remotes, 'pigweed', pw_root_dir):
        append_project(projects, 'pigweed', pw_root_dir)
    if maybe_append_remote(remotes, 'zephyr', zephyr_root_dir):
        append_project(projects, 'zephyr', zephyr_root_dir)

    return {
        'manifest': {
            'remotes': remotes,
            'projects': projects,
        },
    }


def print_manifest() -> None:
    """Print the content of a West manifest for the current Pigweed checkout"""
    _LOG.info(yaml.safe_dump(generate_manifest()))


def main() -> None:
    """Main entry point for the 'pw zephyr' command."""
    parser = get_parser()
    args = parser.parse_args()
    _LOG.addHandler(logging.StreamHandler())
    _LOG.propagate = False
    if args.verbose:
        _LOG.setLevel(logging.DEBUG)
    else:
        _LOG.setLevel(logging.INFO)

    if args.zephyr_subcommand == 'manifest':
        print_manifest()
    else:
        parser.print_usage()


if __name__ == '__main__':
    main()



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tooling to generate C++ constants from a yaml sensor definition."""

import argparse
from dataclasses import dataclass, fields, is_dataclass
from collections.abc import Sequence
import io
import re
import sys
from typing import Type, List, Any
import typing

import yaml


def kid_from_name(name: str) -> str:
    """
    Generate a const style ID name from a given name string. Example:
      If name is "sample_rate", the ID would be kSampleRate

    Args:
      name: the name to convert to an ID
    Returns:
      C++ style 'k' prefixed camel cased ID
    """
    return "k" + ''.join(ele.title() for ele in re.split(r"[\s_\-\,]+", name))


class Printable:
    """Common printable object"""

    def __init__(
        self, item_id: str, name: str, description: str | None
    ) -> None:
        self.id: str = item_id
        self.name: str = name
        self.description: str | None = description

    def __hash__(self) -> int:
        return hash((self.id, self.name, self.description))

    @property
    def variable_name(self) -> str:
        """Convert the 'id' to a constant variable name in C++."""
        return kid_from_name(self.id)

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Baseclass for a printable sensor object

        Args:
          writer: IO writer used to print values.
        """
        writer.write(
            f"""
/// @brief {self.name}
"""
        )
        if self.description:
            writer.write(
                f"""///
/// {self.description}
"""
            )


@dataclass
class UnitsSpec:
    """Typing for the Units definition dictionary."""

    name: str
    symbol: str


class Units(Printable):
    """A single unit representation"""

    symbol: str

    def __init__(self, unit_id: str, definition: UnitsSpec) -> None:
        """
        Create a new unit object

        Args:
          unit_id: The ID of the unit
          definition: A dictionary of the unit definition
        """
        super().__init__(
            item_id=unit_id,
            name=definition.name,
            description=definition.name,
        )
        self.symbol: str = definition.symbol

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.symbol))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Units):
            return False
        return (
            self.id == value.id
            and self.name == value.name
            and self.description == value.description
            and self.symbol == value.symbol
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Print header definition for this unit

        Args:
          writer: IO writer used to print values.
        """
        super().print(writer=writer)
        writer.write(
            f"""PW_SENSOR_UNIT_TYPE(
    {super().variable_name},
    "PW_SENSOR_UNITS_TYPE",
    "{self.name}",
    "{self.symbol}"
);
"""
        )


@dataclass
class AttributeSpec:
    """Typing for the Attribute definition dictionary."""

    name: str
    description: str


class Attribute(Printable):
    """A single attribute representation."""

    def __init__(self, attr_id: str, definition: AttributeSpec) -> None:
        super().__init__(
            item_id=attr_id,
            name=definition.name,
            description=definition.description,
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Print header definition for this attribute

        Args:
          writer: IO writer used to print values.
        """
        super().print(writer=writer)
        writer.write(
            f"""PW_SENSOR_ATTRIBUTE_TYPE(
    {super().variable_name},
    "PW_SENSOR_ATTRIBUTE_TYPE",
    "{self.name}"
);
"""
        )


@dataclass
class ChannelSpec:
    """Typing for the Channel definition dictionary."""

    name: str
    description: str
    units: str


class Channel(Printable):
    """A single channel representation."""

    def __init__(
        self, channel_id: str, definition: ChannelSpec, units: dict[str, Units]
    ) -> None:
        super().__init__(
            item_id=channel_id,
            name=definition.name,
            description=definition.description,
        )
        self.units: Units = units[definition.units]

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.units))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Channel):
            return False
        return (
            self.id == value.id
            and self.name == value.name
            and self.description == value.description
            and self.units == value.units
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Print header definition for this channel

        Args:
          writer: IO writer used to print values.
        """
        super().print(writer=writer)
        writer.write(
            f"""
PW_SENSOR_MEASUREMENT_TYPE({super().variable_name},
                           "PW_SENSOR_MEASUREMENT_TYPE",
                           "{self.name}",
                           ::pw::sensor::units::{self.units.variable_name}
);"""
        )


@dataclass
class TriggerSpec:
    """Typing for the Trigger definition dictionary."""

    name: str
    description: str


class Trigger(Printable):
    """A single trigger representation."""

    def __init__(self, trigger_id: str, definition: TriggerSpec) -> None:
        super().__init__(
            item_id=trigger_id,
            name=definition.name,
            description=definition.description,
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Print header definition for this trigger

        Args:
          writer: IO writer used to print values.
        """
        super().print(writer=writer)
        writer.write(
            f"""PW_SENSOR_TRIGGER_TYPE(
    {super().variable_name},
    "PW_SENSOR_TRIGGER_TYPE",
    "{self.name}"
);
"""
        )


@dataclass
class SensorAttributeSpec:
    """Typing for the SensorAttribute definition dictionary."""

    channel: str
    attribute: str
    units: str


class SensorAttribute(Printable):
    """An attribute instance belonging to a sensor"""

    @staticmethod
    def id_from_definition(definition: SensorAttributeSpec) -> str:
        """
        Get a unique ID for the channel/attribute pair (not sensor specific)

        Args:
          definition: A dictionary of the attribute definition
        Returns:
          String representation for the channel/attribute pair
        """
        return f"{definition.channel}-{definition.attribute}"

    @staticmethod
    def name_from_definition(definition: SensorAttributeSpec) -> str:
        """
        Get a unique name for the channel/attribute pair (not sensor specific)

        Args:
          definition: A dictionary of the attribute definition
        Returns:
          String representation of the human readable name for the
          channel/attribute pair.
        """
        return f"{definition.channel}'s {definition.attribute} attribute"

    @staticmethod
    def description_from_definition(definition: SensorAttributeSpec) -> str:
        """
        Get the description for the channel/attribute pair (not sensor specific)

        Args:
          definition: A dictionary of the attribute definition
        Returns:
          A description string for the channel/attribute pair.
        """
        return (
            f"Allow the configuration of the {definition.channel}'s "
            + f"{definition.attribute} attribute"
        )

    def __init__(self, definition: SensorAttributeSpec) -> None:
        super().__init__(
            item_id=SensorAttribute.id_from_definition(definition=definition),
            name=SensorAttribute.name_from_definition(definition=definition),
            description=SensorAttribute.description_from_definition(
                definition=definition
            ),
        )
        self.attribute: str = definition.attribute
        self.channel: str = definition.channel
        self.units: str = definition.units

    def __hash__(self) -> int:
        return hash(
            (super().__hash__(), self.attribute, self.channel, self.units)
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        super().print(writer)
        writer.write(
            f"""
PW_SENSOR_ATTRIBUTE_INSTANCE({self.variable_name},
                             channels::{kid_from_name(self.channel)},
                             attributes::{kid_from_name(self.attribute)},
                             units::{kid_from_name(self.units)});
"""
        )


@dataclass
class CompatibleSpec:
    """Typing for the Compatible dictionary."""

    org: str
    part: str


@dataclass
class SensorSpec:
    """Typing for the Sensor definition dictionary."""

    description: str
    compatible: CompatibleSpec
    supported_buses: List[str]
    attributes: List[SensorAttributeSpec]
    channels: dict[str, List[ChannelSpec]]
    triggers: List[Any]


class Sensor(Printable):
    """Represent a single sensor type instance"""

    @staticmethod
    def sensor_id_to_name(sensor_id: str) -> str:
        """
        Convert a sensor ID to a human readable name

        Args:
          sensor_id: The ID of the sensor
        Returns:
          Human readable name based on the ID
        """
        return sensor_id.replace(',', ' ')

    def __init__(self, item_id: str, definition: SensorSpec) -> None:
        super().__init__(
            item_id=item_id,
            name=Sensor.sensor_id_to_name(item_id),
            description=definition.description,
        )
        self.compatible_org: str = definition.compatible.org
        self.compatible_part: str = definition.compatible.part
        self.chan_count: int = len(definition.channels)
        self.attr_count: int = len(definition.attributes)
        self.trig_count: int = len(definition.triggers)
        self.attributes: Sequence[SensorAttribute] = [
            SensorAttribute(definition=spec) for spec in definition.attributes
        ]

    @property
    def namespace(self) -> str:
        """
        The namespace which owns the sensor (the org from the compatible)

        Returns:
          The C++ style namespace name of the org.
        """
        return self.compatible_org.replace("-", "_")

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self.compatible_org,
                self.compatible_part,
                self.chan_count,
                self.attr_count,
                self.trig_count,
            )
        )

    def print(self, writer: io.TextIOWrapper) -> None:
        """
        Print header definition for this trigger

        Args:
          writer: IO writer used to print values.
        """
        writer.write(
            f"""
namespace {self.namespace} {{
class {self.compatible_part.upper()}
  : public pw::sensor::zephyr::ZephyrSensor<{len(self.attributes)}> {{
 public:
  {self.compatible_part.upper()}(const struct device* dev)
      : pw::sensor::zephyr::ZephyrSensor<{len(self.attributes)}>(
            dev,
            {{"""
        )
        for attribute in self.attributes:
            writer.write(
                f"""
             Attribute::Build<{attribute.variable_name}>(),"""
            )
        writer.write(
            f"""
            }}) {{}}
}};
}}  // namespace {self.namespace}
"""
        )


@dataclass
class Args:
    """CLI arguments"""

    package: Sequence[str]
    language: str
    zephyr: bool


class CppHeader:
    """Generator for a C++ header"""

    def __init__(
        self,
        using_zephyr: bool,
        package: Sequence[str],
        attributes: Sequence[Attribute],
        channels: Sequence[Channel],
        triggers: Sequence[Trigger],
        units: Sequence[Units],
        sensors: Sequence[Sensor],
    ) -> None:
        """
        Args:
          package: The package name used in the output. In C++ we'll convert
            this to a namespace.
          attributes: A sequence of attributes which will be exposed in the
            'attributes' namespace
          channels: A sequence of channels which will be exposed in the
            'channels' namespace
          triggers: A sequence of triggers which will be exposed in the
            'triggers' namespace
          units: A sequence of units which should be exposed in the 'units'
            namespace
        """
        self._using_zephyr: bool = using_zephyr
        self._package: str = '::'.join(package)
        self._attributes: Sequence[Attribute] = attributes
        self._channels: Sequence[Channel] = channels
        self._triggers: Sequence[Trigger] = triggers
        self._units: Sequence[Units] = units
        self._sensors: Sequence[Sensor] = sensors
        self._sensor_attributes: set[SensorAttribute] = set()
        for sensor in sensors:
            self._sensor_attributes.update(sensor.attributes)

    def __str__(self) -> str:
        writer = io.StringIO()
        self._print_header(writer=writer)
        self._print_constants(writer=writer)
        self._print_footer(writer=writer)
        return writer.getvalue()

    def _print_header(self, writer: io.TextIOWrapper) -> None:
        """
        Print the top part of the .h file (pragma, includes, and namespace)

        Args:
          writer: Where to write the text to
        """
        writer.write(
            "/* Auto-generated file, do not edit */\n"
            "#pragma once\n"
            "\n"
            "#include \"pw_sensor/types.h\"\n"
        )
        if self._package:
            writer.write(f"namespace {self._package} {{\n\n")

    def _print_constants(self, writer: io.TextIOWrapper) -> None:
        """
        Print the constants definitions from self._attributes, self._channels,
        and self._trigger

        Args:
            writer: Where to write the text
        """

        self._print_in_namespace(
            namespace="units", printables=self._units, writer=writer
        )
        self._print_in_namespace(
            namespace="attributes", printables=self._attributes, writer=writer
        )
        self._print_in_namespace(
            namespace="channels", printables=self._channels, writer=writer
        )
        self._print_in_namespace(
            namespace="triggers", printables=self._triggers, writer=writer
        )
        for sensor_attribute in self._sensor_attributes:
            sensor_attribute.print(writer=writer)

    @staticmethod
    def _print_in_namespace(
        namespace: str,
        printables: Sequence[Printable],
        writer: io.TextIOWrapper,
    ) -> None:
        """
        Print constants definitions wrapped in a namespace
        Args:
          namespace: The namespace to use
          printables: A sequence of printable objects
          writer: Where to write the text
        """
        writer.write(f"\nnamespace {namespace} {{\n")
        for printable in printables:
            printable.print(writer=writer)
        writer.write(f"\n}}  // namespace {namespace}\n")

    def _print_footer(self, writer: io.TextIOWrapper) -> None:
        """
        Write the bottom part of the .h file (closing namespace)

        Args:
            writer: Where to write the text
        """
        if self._package:
            writer.write(f"\n}}  // namespace {self._package}")

        if self._using_zephyr:
            self._print_zephyr_mapping(writer=writer)

    def _print_zephyr_mapping(self, writer: io.TextIOWrapper) -> None:
        """
        Generate Zephyr type maps for channels, attributes, and triggers.

        Args:
            writer: Where to write the text
        """
        writer.write(
            f"""
#include <zephyr/generated/sensor_constants.h>
#include \"pw_containers/flat_map.h\"

namespace pw::sensor::zephyr {{

class ZephyrAttributeMap
    : public pw::containers::FlatMap<uint32_t, uint32_t,
                                     {len(self._attributes)}> {{
 public:
  constexpr ZephyrAttributeMap()
      : pw::containers::FlatMap<uint32_t, uint32_t,
                                {len(self._attributes)}>({{{{"""
        )
        for attribute in self._attributes:
            attribute_type = (
                f"{self._package}::attributes::"
                + f"{attribute.variable_name}::kAttributeType"
            )
            writer.write(
                f"""
            {{{attribute_type},
             SENSOR_ATTR_{attribute.id.upper()}}},"""
            )
        writer.write("\n      }}) {}\n};")
        writer.write(
            f"""

class ZephyrChannelMap
    : public pw::containers::FlatMap<uint32_t, uint32_t,
                                     {len(self._channels)}> {{
 public:
  constexpr ZephyrChannelMap()
      : pw::containers::FlatMap<uint32_t, uint32_t,
                                {len(self._channels)}>({{{{"""
        )
        for channel in self._channels:
            measurement_name = (
                f"{self._package}::channels::"
                + f"{channel.variable_name}::kMeasurementName"
            )
            writer.write(
                f"""
            {{{measurement_name},
             SENSOR_CHAN_{channel.id.upper()}}},"""
            )
        writer.write(
            """
      }}) {}
};

extern ZephyrAttributeMap kAttributeMap;
extern ZephyrChannelMap kChannelMap;

}  // namespace pw::sensor::zephyr

#include "pw_sensor_zephyr/sensor.h"
namespace pw::sensor {
"""
        )
        for sensor in self._sensors:
            sensor.print(writer=writer)
        writer.write(
            """
}  // namespace pw::sensor
"""
        )


@dataclass
class InputSpec:
    """Typing for the InputData spec dictionary"""

    units: dict[str, UnitsSpec]
    attributes: dict[str, AttributeSpec]
    channels: dict[str, ChannelSpec]
    triggers: dict[str, TriggerSpec]
    sensors: dict[str, SensorSpec]


class InputData:
    """
    Wrapper class for all the input data parsed out into: Units, Attribute,
    Channel, Trigger, and Sensor types (or sub-types).
    """

    def __init__(
        self,
        spec: InputSpec,
        units_type: Type[Units] = Units,
        attribute_type: Type[Attribute] = Attribute,
        channel_type: Type[Channel] = Channel,
        trigger_type: Type[Trigger] = Trigger,
        sensor_type: Type[Sensor] = Sensor,
    ) -> None:
        """
        Parse the input spec and create all the input data types.

        Args:
          spec: The input spec dictionary
          units_type: The type to use for units
          attribute_type: The type to use for attributes
          channel_type: The type to use for channels
          trigger_type: The type to use for triggers
          sensor_type: The type to use for sensors
        """
        self.all_attributes: set[Attribute] = set()
        self.all_channels: set[Channel] = set()
        self.all_triggers: set[Trigger] = set()
        self.all_units: dict[str, Units] = {}
        self.all_sensors: set[Sensor] = set()
        for units_id, units_spec in spec.units.items():
            units = units_type(unit_id=units_id, definition=units_spec)
            assert units not in self.all_units.values()
            self.all_units[units_id] = units
        for attribute_id, attribute_spec in spec.attributes.items():
            attribute = attribute_type(
                attr_id=attribute_id, definition=attribute_spec
            )
            assert not attribute in self.all_attributes
            self.all_attributes.add(attribute)
        for channel_id, channel_spec in spec.channels.items():
            channel = channel_type(
                channel_id=channel_id,
                definition=channel_spec,
                units=self.all_units,
            )
            assert not channel in self.all_channels
            self.all_channels.add(channel)
        for trigger_id, trigger_spec in spec.triggers.items():
            trigger = trigger_type(
                trigger_id=trigger_id, definition=trigger_spec
            )
            assert not trigger in self.all_triggers
            self.all_triggers.add(trigger)
        for sensor_id, sensor_spec in spec.sensors.items():
            sensor = sensor_type(item_id=sensor_id, definition=sensor_spec)
            assert not sensor in self.all_sensors
            self.all_sensors.add(sensor)


def is_list_type(t) -> bool:
    """
    Checks if the given type `t` is either a list or typing.List.

    Args:
        t: The type to check.
    Returns:
        True if `t` is a list type, False otherwise.
    """
    origin = typing.get_origin(t)
    return origin is list or (origin is list and typing.get_args(t) == ())


def is_primitive(value):
    """Checks if the given value is of a primitive type.

    Args:
        value: The value to check.

    Returns:
        True if the value is of a primitive type, False otherwise.
    """
    return isinstance(value, (int, float, complex, str, bool))


def create_dataclass_from_dict(cls, data, indent: int = 0):
    """Recursively creates a dataclass instance from a nested dictionary."""

    field_values = {}

    if is_list_type(cls):
        result = []
        for item in data:
            result.append(
                create_dataclass_from_dict(
                    typing.get_args(cls)[0], item, indent + 2
                )
            )
        return result

    if is_primitive(data):
        return data

    for field in fields(cls):
        field_value = data.get(field.name)
        if field_value is None:
            field_value = data.get(field.name.replace('_', '-'))

        assert field_value is not None

        # We need to check if the field is a List, dictionary, or another
        # dataclass. If it is, recurse.
        if is_list_type(field.type):
            item_type = typing.get_args(field.type)[0]
            print((" " * indent) + str(item_type), file=sys.stderr)
            field_value = [
                create_dataclass_from_dict(item_type, item, indent + 2)
                for item in field_value
            ]
        elif dict in field.type.__mro__:
            value_type = typing.get_args(field.type)[1]
            field_value = {
                key: create_dataclass_from_dict(value_type, val, indent + 2)
                for key, val in field_value.items()
            }
        elif is_dataclass(field.type):
            field_value = create_dataclass_from_dict(
                field.type, field_value, indent + 2
            )

        field_values[field.name] = field_value

    return cls(**field_values)


def main() -> None:
    """
    Main entry point, this function will:
    - Get CLI flags
    - Read YAML from stdin
    - Find all attribute, channel, trigger, and unit definitions
    - Print header
    """
    args = get_args()
    yaml_input = yaml.safe_load(sys.stdin)
    spec: InputSpec = create_dataclass_from_dict(InputSpec, yaml_input)
    data = InputData(spec=spec)

    if args.language == "cpp":
        out = CppHeader(
            using_zephyr=args.zephyr,
            package=args.package,
            attributes=list(data.all_attributes),
            channels=list(data.all_channels),
            triggers=list(data.all_triggers),
            units=list(data.all_units.values()),
            sensors=list(data.all_sensors),
        )
    else:
        raise ValueError(f"Invalid language selected: '{args.language}'")
    print(out)


def validate_package_arg(value: str) -> str:
    """
    Validate that the package argument is a valid string

    Args:
      value: The package name

    Returns:
      The same value after being validated.
    """
    if value is None or value == "":
        return value
    if not re.match(r"[a-zA-Z_$][\w$]*(\.[a-zA-Z_$][\w$]*)*", value):
        raise argparse.ArgumentError(
            argument=None,
            message=f"Invalid string {value}. Must use alphanumeric values "
            "separated by dots.",
        )
    return value


def get_args() -> Args:
    """
    Get CLI arguments

    Returns:
      Typed arguments class instance
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--package",
        "-pkg",
        default="",
        type=validate_package_arg,
        help="Output package name separated by '.', example: com.google",
    )
    parser.add_argument(
        "--language",
        type=str,
        choices=["cpp"],
        default="cpp",
    )
    parser.add_argument(
        "--zephyr",
        action="store_true",
    )
    args = parser.parse_args()
    return Args(
        package=args.package.split("."),
        language=args.language,
        zephyr=args.zephyr,
    )


if __name__ == "__main__":
    main()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Configure C/C++ IDE support for Pigweed projects.

We support C/C++ code analysis via ``clangd``, or other language servers that
are compatible with the ``clangd`` compilation database format.

While clangd can work well out of the box for typical C++ codebases, some work
is required to coax it to work for embedded projects. In particular, Pigweed
projects use multiple toolchains within a distinct environment, and almost
always define multiple targets. This means compilation units are likely have
multiple compile commands and the toolchain executables are unlikely to be in
your path. ``clangd`` is not equipped to deal with this out of the box. We
handle this by:

- Processing the compilation database produced by the build system into
  multiple internally-consistent compilation databases, one for each target
  (where a "target" is a particular build for a particular system using a
  particular toolchain).

- Creating unambiguous paths to toolchain drivers to ensure the right toolchain
  is used and that clangd knows where to find that toolchain's system headers.

- Providing tools for working with several compilation databases that are
  spiritually similar to tools like ``pyenv``, ``rbenv``, etc.

In short, we take the probably-broken compilation database that the build system
generates, process it into several not-broken compilation databases in the
``pw_ide`` working directory, and provide a stable symlink that points to the
selected active target's compliation database. If ``clangd`` is configured to
point at the symlink and is set up with the right paths, you'll get code
intelligence.
"""

from __future__ import annotations

from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
import functools
import glob
from hashlib import sha1
from io import TextIOBase
import json
import logging
from pathlib import Path
import platform
import random
import re
import sys
from typing import (
    Any,
    cast,
    Generator,
    Iterator,
    TypedDict,
)

from pw_cli.env import pigweed_environment

from pw_ide.exceptions import (
    BadCompDbException,
    InvalidTargetException,
    MissingCompDbException,
    UnresolvablePathException,
)

from pw_ide.settings import PigweedIdeSettings
from pw_ide.symlinks import set_symlink

_LOG = logging.getLogger(__package__)
env = pigweed_environment()

COMPDB_FILE_NAME = 'compile_commands.json'
STABLE_CLANGD_DIR_NAME = '.stable'
_CPP_IDE_FEATURES_DATA_FILE = 'pw_ide_state.json'
_UNSUPPORTED_TOOLCHAIN_EXECUTABLES = ('_pw_invalid', 'python')
_SUPPORTED_WRAPPER_EXECUTABLES = ('ccache',)


@dataclass(frozen=True)
class CppIdeFeaturesTarget:
    """Data pertaining to a C++ code analysis target."""

    name: str
    compdb_file_path: Path
    num_commands: int
    is_enabled: bool = True

    def __str__(self) -> str:
        return self.name

    def serialized(self) -> dict[str, Any]:
        return {
            **asdict(self),
            **{
                'compdb_file_path': str(self.compdb_file_path),
            },
        }

    @classmethod
    def deserialize(cls, **data) -> CppIdeFeaturesTarget:
        return cls(
            **{
                **data,
                **{
                    'compdb_file_path': Path(data['compdb_file_path']),
                },
            }
        )


CppCompilationDatabaseFileHashes = dict[Path, str]
CppCompilationDatabaseFileTargets = dict[Path, list[CppIdeFeaturesTarget]]


@dataclass
class CppIdeFeaturesData:
    """State data about C++ code analysis features."""

    targets: dict[str, CppIdeFeaturesTarget] = field(default_factory=dict)
    current_target: CppIdeFeaturesTarget | None = None
    compdb_hashes: CppCompilationDatabaseFileHashes = field(
        default_factory=dict
    )
    compdb_targets: CppCompilationDatabaseFileTargets = field(
        default_factory=dict
    )

    def serialized(self) -> dict[str, Any]:
        return {
            'current_target': self.current_target.serialized()
            if self.current_target is not None
            else None,
            'targets': {
                name: target_data.serialized()
                for name, target_data in self.targets.items()
            },
            'compdb_hashes': {
                str(path): hash_str
                for path, hash_str in self.compdb_hashes.items()
            },
            'compdb_targets': {
                str(path): [
                    target_data.serialized() for target_data in target_data_list
                ]
                for path, target_data_list in self.compdb_targets.items()
            },
        }

    @classmethod
    def deserialize(cls, **data) -> CppIdeFeaturesData:
        return cls(
            current_target=CppIdeFeaturesTarget.deserialize(
                **data['current_target']
            )
            if data['current_target'] is not None
            else None,
            targets={
                name: CppIdeFeaturesTarget.deserialize(**target_data)
                for name, target_data in data['targets'].items()
            },
            compdb_hashes={
                Path(path_str): hash_str
                for path_str, hash_str in data['compdb_hashes'].items()
            },
            compdb_targets={
                Path(path_str): [
                    CppIdeFeaturesTarget.deserialize(**target_data)
                    for target_data in target_data_list
                ]
                for path_str, target_data_list in data['compdb_targets'].items()
            },
        )


class CppIdeFeaturesState:
    """Container for IDE features state data."""

    def __init__(self, pw_ide_settings: PigweedIdeSettings) -> None:
        self.settings = pw_ide_settings

    def __len__(self) -> int:
        return len(self.targets)

    def __getitem__(self, index: str) -> CppIdeFeaturesTarget:
        return self.targets[index]

    def __iter__(self) -> Generator[CppIdeFeaturesTarget, None, None]:
        return (target for target in self.targets.values())

    @property
    def stable_target_link(self) -> Path:
        return self.settings.working_dir / STABLE_CLANGD_DIR_NAME

    @contextmanager
    def _file(self) -> Generator[CppIdeFeaturesData, None, None]:
        """A simple key-value store for state data."""
        file_path = self.settings.working_dir / _CPP_IDE_FEATURES_DATA_FILE

        try:
            with open(file_path) as file:
                data = CppIdeFeaturesData.deserialize(**json.load(file))
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            data = CppIdeFeaturesData()

        yield data

        with open(file_path, 'w') as file:
            json.dump(data.serialized(), file, indent=2)

    @property
    def targets(self) -> dict[str, CppIdeFeaturesTarget]:
        with self._file() as state:
            exclude_predicate = (
                lambda x: x not in self.settings.targets_exclude
                if len(self.settings.targets_exclude) > 0
                else lambda x: True
            )
            include_predicate = (
                lambda x: x in self.settings.targets_include
                if len(self.settings.targets_include) > 0
                else lambda x: True
            )

            return {
                name: target
                for (name, target) in state.targets.items()
                if exclude_predicate(name) and include_predicate(name)
            }

    @targets.setter
    def targets(self, new_targets: dict[str, CppIdeFeaturesTarget]) -> None:
        with self._file() as state:
            state.targets = new_targets

    @property
    def current_target(self) -> CppIdeFeaturesTarget | None:
        with self._file() as state:
            return state.current_target

    @current_target.setter
    def current_target(
        self, new_current_target: str | CppIdeFeaturesTarget | None
    ) -> None:
        with self._file() as state:
            if new_current_target is None:
                state.current_target = None
            else:
                if isinstance(new_current_target, CppIdeFeaturesTarget):
                    name = new_current_target.name
                    new_current_target_inst = new_current_target
                else:
                    name = new_current_target

                    try:
                        new_current_target_inst = state.targets[name]
                    except KeyError:
                        raise InvalidTargetException

                if not new_current_target_inst.compdb_file_path.exists():
                    raise MissingCompDbException

                set_symlink(
                    new_current_target_inst.compdb_file_path.parent,
                    self.stable_target_link,
                )

                state.current_target = state.targets[name]

    @property
    def max_commands_target(self) -> CppIdeFeaturesTarget | None:
        with self._file() as state:
            if len(state.targets) == 0:
                return None

            max_commands_target_name = sorted(
                [
                    (name, target.num_commands)
                    for name, target in state.targets.items()
                ],
                key=lambda x: x[1],
                reverse=True,
            )[0][0]

            return state.targets[max_commands_target_name]

    @property
    def compdb_hashes(self) -> CppCompilationDatabaseFileHashes:
        with self._file() as state:
            return state.compdb_hashes

    @compdb_hashes.setter
    def compdb_hashes(
        self, new_compdb_hashes: CppCompilationDatabaseFileHashes
    ) -> None:
        with self._file() as state:
            state.compdb_hashes = new_compdb_hashes

    @property
    def compdb_targets(self) -> CppCompilationDatabaseFileTargets:
        with self._file() as state:
            return state.compdb_targets

    @compdb_targets.setter
    def compdb_targets(
        self, new_compdb_targets: CppCompilationDatabaseFileTargets
    ) -> None:
        with self._file() as state:
            state.compdb_targets = new_compdb_targets


def path_to_executable(
    exe: str,
    *,
    default_path: Path | None = None,
    path_globs: list[str] | None = None,
    strict: bool = False,
) -> Path | None:
    """Return the path to a compiler executable.

    In a ``clang`` compile command, the executable may or may not include a
    path. For example:

    .. code-block:: none

       /usr/bin/clang      <- includes a path
       ../path/to/my/clang <- includes a path
       clang               <- doesn't include a path

    If it includes a path, then ``clangd`` will have no problem finding the
    driver, so we can simply return the path. If the executable *doesn't*
    include a path, then ``clangd`` will search ``$PATH``, and may not find the
    intended driver unless you actually want the default system toolchain or
    Pigweed paths have been added to ``$PATH``. So this function provides two
    options for resolving those ambiguous paths:

    - Provide a default path, and all executables without a path will be
      re-written with a path within the default path.
    - Provide the a set of globs that will be used to search for the executable,
      which will normally be the query driver globs used with clangd.

    By default, if neither of these options is chosen, or if the executable
    cannot be found within the provided globs, the pathless executable that was
    provided will be returned, and clangd will resort to searching $PATH. If you
    instead pass ``strict=True``, this will raise an exception if an unambiguous
    path cannot be constructed.

    This function only tries to ensure that all executables have a path to
    eliminate ambiguity. A couple of important things to keep in mind:

    - This doesn't guarantee that the path exists or an executable actually
      exists at the path. It only ensures that some path is provided to an
      executable.
    - An executable being present at the indicated path doesn't guarantee that
      it will work flawlessly for clangd code analysis. The clangd
      ``--query-driver`` argument needs to include a path to this executable in
      order for its bundled headers to be resolved correctly.

    This function also filters out invalid or unsupported drivers. For example,
    build systems will sometimes naively include build steps for Python or other
    languages in the compilation database, which are not usable with clangd.
    As a result, this function has four possible end states:

    - It returns a path with an executable that can be used as a ``clangd``
      driver.
    - It returns ``None``, meaning the compile command was invalid.
    - It returns the same string that was provided (as a ``Path``), if a path
      couldn't be resolved and ``strict=False``.
    - It raises an ``UnresolvablePathException`` if the executable cannot be
      placed in an unambiguous path and ``strict=True``.
    """
    maybe_path = Path(exe)

    # We were give an empty string, not a path. Not a valid command.
    if len(maybe_path.parts) == 0:
        _LOG.debug("Invalid executable path. The path was an empty string.")
        return None

    # Determine if the executable name matches unsupported drivers.
    is_supported_driver = True

    for unsupported_executable in _UNSUPPORTED_TOOLCHAIN_EXECUTABLES:
        if unsupported_executable in maybe_path.name:
            is_supported_driver = False

    if not is_supported_driver:
        _LOG.debug(
            "Invalid executable path. This is not a supported driver: %s", exe
        )
        return None

    # Now, ensure the executable has a path.

    # This is either a relative or absolute path -- return it.
    if len(maybe_path.parts) > 1:
        return maybe_path

    # If we got here, there's only one "part", so we assume it's an executable
    # without a path. This logic doesn't work with a path like `./exe` since
    # that also yields only one part. So currently this breaks if you actually
    # have your compiler executable in your root build directory, which is
    # (hopefully) very rare.

    # If we got a default path, use it.
    if default_path is not None:
        return default_path / maybe_path

    # Otherwise, try to find the executable within the query driver globs.
    # Note that unlike the previous paths, this path will only succeed if an
    # executable actually exists somewhere in the query driver globs.
    if path_globs is not None:
        for path_glob in path_globs:
            for path_str in glob.iglob(path_glob):
                path = Path(path_str)
                if path.name == maybe_path.name:
                    return path.absolute()

    if strict:
        raise UnresolvablePathException(
            f'Cannot place {exe} in an unambiguous path!'
        )

    return maybe_path


def command_parts(command: str) -> tuple[str | None, str, list[str]]:
    """Return the executable string and the rest of the command tokens.

    If the command contains a prefixed wrapper like `ccache`, it will be
    extracted separately. So the return value contains:
        (wrapper, compiler executable, all other tokens)
    """
    parts = command.split()
    curr = ''
    wrapper = None

    try:
        curr = parts.pop(0)
    except IndexError:
        return (None, curr, [])

    if curr in _SUPPORTED_WRAPPER_EXECUTABLES:
        wrapper = curr

        while curr := parts.pop(0):
            # This is very `ccache`-centric. It will work for other wrappers
            # that use KEY=VALUE-style options or no options at all, but will
            # not work for other cases.
            if re.fullmatch(r'(.*)=(.*)', curr):
                wrapper = f'{wrapper} {curr}'
            else:
                break

    return (wrapper, curr, parts)


# This is a clumsy way to express optional keys, which is not directly
# supported in TypedDicts right now.
# TODO(chadnorvell): Use `NotRequired` when we support Python 3.11.
class BaseCppCompileCommandDict(TypedDict):
    file: str
    directory: str
    output: str | None


class CppCompileCommandDictWithCommand(BaseCppCompileCommandDict):
    command: str


class CppCompileCommandDictWithArguments(BaseCppCompileCommandDict):
    arguments: list[str]


CppCompileCommandDict = (
    CppCompileCommandDictWithCommand | CppCompileCommandDictWithArguments
)


class CppCompileCommand:
    """A representation of a clang compilation database compile command.

    See: https://clang.llvm.org/docs/JSONCompilationDatabase.html
    """

    def __init__(
        self,
        file: str,
        directory: str,
        command: str | None = None,
        arguments: list[str] | None = None,
        output: str | None = None,
    ) -> None:
        # Per the spec, either one of these two must be present. clangd seems
        # to prefer "arguments" when both are present.
        if command is None and arguments is None:
            raise TypeError(
                'A compile command requires either \'command\' '
                'or \'arguments\'.'
            )

        if command is None:
            raise TypeError(
                'Compile commands without \'command\' ' 'are not supported yet.'
            )

        self._command = command
        self._arguments = arguments
        self._file = file
        self._directory = directory

        _, executable, tokens = command_parts(command)
        self._executable_path = Path(executable)
        self._inferred_output: str | None = None

        try:
            # Find the output argument and grab its value.
            output_flag_idx = tokens.index('-o')
            self._inferred_output = tokens[output_flag_idx + 1]
        except ValueError:
            # No -o found, probably not a C/C++ compile command.
            self._inferred_output = None
        except IndexError:
            # It has an -o but no argument after it.
            raise TypeError(
                'Failed to load compile command with no output argument!'
            )

        self._provided_output = output
        self.target: str | None = None

    @property
    def file(self) -> str:
        return self._file

    @property
    def directory(self) -> str:
        return self._directory

    @property
    def command(self) -> str | None:
        return self._command

    @property
    def arguments(self) -> list[str] | None:
        return self._arguments

    @property
    def output(self) -> str | None:
        # We're ignoring provided output values for now.
        return self._inferred_output

    @property
    def output_path(self) -> Path | None:
        if self.output is None:
            return None

        return Path(self.directory) / Path(self.output)

    @property
    def executable_path(self) -> Path:
        return self._executable_path

    @property
    def executable_name(self) -> str:
        return self.executable_path.name

    @classmethod
    def from_dict(
        cls, compile_command_dict: dict[str, Any]
    ) -> CppCompileCommand:
        return cls(
            # We want to let possible Nones through to raise at runtime.
            file=cast(str, compile_command_dict.get('file')),
            directory=cast(str, compile_command_dict.get('directory')),
            command=compile_command_dict.get('command'),
            arguments=compile_command_dict.get('arguments'),
            output=compile_command_dict.get('output'),
        )

    @classmethod
    def try_from_dict(
        cls, compile_command_dict: dict[str, Any]
    ) -> CppCompileCommand | None:
        try:
            return cls.from_dict(compile_command_dict)
        except TypeError:
            return None

    def process(
        self,
        *,
        default_path: Path | None = None,
        path_globs: list[str] | None = None,
        strict: bool = False,
    ) -> CppCompileCommand | None:
        """Process a compile command.

        At minimum, a compile command from a clang compilation database needs to
        be correlated with its target, and this method returns the target name
        with the compile command. But it also cleans up other things we need for
        reliable code intelligence:

        - Some targets may not be valid C/C++ compile commands. For example,
          some build systems will naively include build steps for Python or for
          linting commands. We want to filter those out.

        - Some compile commands don't provide a path to the compiler executable
          (referred to by clang as the "driver"). In that case, clangd is very
          unlikely to find the executable unless it happens to be in ``$PATH``.
          The ``--query-driver`` argument to ``clangd`` allowlists
          executables/drivers for use its use, but clangd doesn't use it to
          resolve ambiguous paths. We bridge that gap here. Any executable
          without a path will be either placed in the provided default path or
          searched for in the query driver globs and be replaced with a path to
          the executable.
        """
        if self.command is None:
            raise NotImplementedError(
                'Compile commands without \'command\' ' 'are not supported yet.'
            )

        wrapper, executable_str, tokens = command_parts(self.command)
        executable_path = path_to_executable(
            executable_str,
            default_path=default_path,
            path_globs=path_globs,
            strict=strict,
        )

        if executable_path is None:
            _LOG.debug(
                "Compile command rejected due to bad executable path: %s",
                self.command,
            )
            return None

        if self.output is None:
            _LOG.debug(
                "Compile command rejected due to no output property: %s",
                self.command,
            )
            return None

        # TODO(chadnorvell): Some commands include the executable multiple
        # times. It's not clear if that affects clangd.
        new_command = f'{str(executable_path)} {" ".join(tokens)}'

        if wrapper is not None:
            new_command = f'{wrapper} {new_command}'

        return self.__class__(
            file=self.file,
            directory=self.directory,
            command=new_command,
            arguments=None,
            output=self.output,
        )

    def as_dict(self) -> CppCompileCommandDict:
        base_compile_command_dict: BaseCppCompileCommandDict = {
            'file': self.file,
            'directory': self.directory,
            'output': self.output,
        }

        # TODO(chadnorvell): Support "arguments". The spec requires that a
        # We don't support "arguments" at all right now. When we do, we should
        # preferentially include "arguments" only, and only include "command"
        # when "arguments" is not present.
        if self.command is not None:
            compile_command_dict: CppCompileCommandDictWithCommand = {
                'command': self.command,
                # Unfortunately dict spreading doesn't work with mypy.
                'file': base_compile_command_dict['file'],
                'directory': base_compile_command_dict['directory'],
                'output': base_compile_command_dict['output'],
            }
        else:
            raise NotImplementedError(
                'Compile commands without \'command\' ' 'are not supported yet.'
            )

        return compile_command_dict


def _path_nearest_parent(path1: Path, path2: Path) -> Path:
    """Get the closest common parent of two paths."""
    # This is the Python < 3.9 version of: if path2.is_relative_to(path1)
    try:
        path2.relative_to(path1)
        return path1
    except ValueError:
        pass

    if path1 == path2:
        return path1

    if len(path1.parts) > len(path2.parts):
        return _path_nearest_parent(path1.parent, path2)

    if len(path1.parts) < len(path2.parts):
        return _path_nearest_parent(path1, path2.parent)

    return _path_nearest_parent(path1.parent, path2.parent)


def _infer_target_pos(target_glob: str) -> list[int]:
    """Infer the position of the target in a compilation unit artifact path."""
    tokens = Path(target_glob).parts
    positions = []

    for pos, token in enumerate(tokens):
        if token == '?':
            positions.append(pos)
        elif token == '*':
            pass
        else:
            raise ValueError(f'Invalid target inference token: {token}')

    return positions


def infer_target(target_glob: str, root: Path, output_path: Path) -> str | None:
    """Infer a target from a compilation unit artifact path.

    See the documentation for ``PigweedIdeSettings.target_inference``."""
    target_pos = _infer_target_pos(target_glob)

    if len(target_pos) == 0:
        return None

    # Depending on the build system and project configuration, the target name
    # may be in the "directory" or the "output" of the compile command. So we
    # need to construct the full path that combines both and use that to search
    # for the target.
    try:
        # The path used for target inference is the path relative to the root
        # dir. If this artifact is a direct child of the root, this just
        # truncates the root off of its path.
        subpath = output_path.relative_to(root)
    except ValueError:
        # If the output path isn't a child path of the root dir, find the
        # closest shared parent dir and use that as the root for truncation.
        common_parent = _path_nearest_parent(root, output_path)
        subpath = output_path.relative_to(common_parent)

    return '_'.join([subpath.parts[pos] for pos in target_pos])


LoadableToCppCompilationDatabase = (
    list[dict[str, Any]] | str | TextIOBase | Path
)


class CppCompilationDatabase:
    """A representation of a clang compilation database.

    See: https://clang.llvm.org/docs/JSONCompilationDatabase.html
    """

    def __init__(
        self,
        root_dir: Path | None = None,
        file_path: Path | None = None,
        source_file_path: Path | None = None,
        target_inference: str | None = None,
    ) -> None:
        self._db: list[CppCompileCommand] = []
        self.file_path: Path | None = file_path
        self.source_file_path: Path | None = source_file_path
        self.source_file_hash: str | None = None

        if target_inference is None:
            self.target_inference = PigweedIdeSettings().target_inference
        else:
            self.target_inference = target_inference

        # Only compilation databases that are loaded will have this, and it
        # contains the root directory of the build that the compilation
        # database is based on. Processed compilation databases will not have
        # a value here.
        self._root_dir = root_dir

    def __len__(self) -> int:
        return len(self._db)

    def __getitem__(self, index: int) -> CppCompileCommand:
        return self._db[index]

    def __iter__(self) -> Generator[CppCompileCommand, None, None]:
        return (compile_command for compile_command in self._db)

    @property
    def file_hash(self) -> str:
        # If this compilation database did not originate from a file, return a
        # hash that is almost certainly not going to match any other hash; these
        # sources are not persistent, so they cannot be compared.
        if self.file_path is None:
            return '%032x' % random.getrandbits(160)

        data = self.file_path.read_text().encode('utf-8')
        return sha1(data).hexdigest()

    def add(self, *commands: CppCompileCommand):
        """Add compile commands to the compilation database."""
        self._db.extend(commands)

    def merge(self, other: CppCompilationDatabase) -> None:
        """Merge values from another database into this one.

        This will not overwrite a compile command that already exists for a
        particular file.
        """
        self_dict = {c.file: c for c in self._db}

        for compile_command in other:
            if compile_command.file not in self_dict:
                self_dict[compile_command.file] = compile_command

        self._db = list(self_dict.values())

    def as_dicts(self) -> list[CppCompileCommandDict]:
        return [compile_command.as_dict() for compile_command in self._db]

    def to_json(self) -> str:
        """Output the compilation database to a JSON string."""

        return json.dumps(self.as_dicts(), indent=2, sort_keys=True)

    def to_file(self, path: Path):
        """Write the compilation database to a JSON file."""
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, 'w') as file:
            json.dump(self.as_dicts(), file, indent=2, sort_keys=True)

    @classmethod
    def load(
        cls,
        compdb_to_load: LoadableToCppCompilationDatabase,
        root_dir: Path,
        target_inference: str | None = None,
    ) -> CppCompilationDatabase:
        """Load a compilation database.

        You can provide a JSON file handle or path, a JSON string, or a native
        Python data structure that matches the format (list of dicts).
        """
        db_as_dicts: list[dict[str, Any]]
        file_path = None

        if isinstance(compdb_to_load, list):
            # The provided data is already in the format we want it to be in,
            # probably, and if it isn't we'll find out when we try to
            # instantiate the database.
            db_as_dicts = compdb_to_load
        else:
            if isinstance(compdb_to_load, Path):
                # The provided data is a path to a file, presumably JSON.
                try:
                    file_path = compdb_to_load
                    compdb_data = compdb_to_load.read_text()
                except FileNotFoundError:
                    raise MissingCompDbException()
            elif isinstance(compdb_to_load, TextIOBase):
                # The provided data is a file handle, presumably JSON.
                file_path = Path(compdb_to_load.name)  # type: ignore
                compdb_data = compdb_to_load.read()
            elif isinstance(compdb_to_load, str):
                # The provided data is a a string, presumably JSON.
                compdb_data = compdb_to_load

            db_as_dicts = json.loads(compdb_data)

        compdb = cls(
            root_dir=root_dir,
            file_path=file_path,
            target_inference=target_inference,
        )

        try:
            compdb.add(
                *[
                    compile_command
                    for compile_command_dict in db_as_dicts
                    if (
                        compile_command := CppCompileCommand.try_from_dict(
                            compile_command_dict
                        )
                    )
                    is not None
                ]
            )
        except TypeError:
            # This will arise if db_as_dicts is not actually a list of dicts
            raise BadCompDbException()

        return compdb

    def process(
        self,
        settings: PigweedIdeSettings,
        *,
        default_path: Path | None = None,
        path_globs: list[str] | None = None,
        strict: bool = False,
        always_output_new: bool = False,
    ) -> CppCompilationDatabasesMap | None:
        """Process a ``clangd`` compilation database file.

        Given a clang compilation database that may have commands for multiple
        valid or invalid targets/toolchains, keep only the valid compile
        commands and store them in target-specific compilation databases.

        If this finds that the processed file is functionally identical to the
        input file (meaning that the input file did not require processing to
        be used successfully with ``clangd``), then it will return ``None``,
        indicating that the original file should be used. This behavior can be
        overridden by setting ``always_output_new``, which will ensure that a
        new compilation database is always written to the working directory and
        original compilation databases outside the working directory are never
        made available for code intelligence.
        """
        if self._root_dir is None:
            raise ValueError(
                'Can only process a compilation database that '
                'contains a root build directory, usually '
                'specified when loading the file. Are you '
                'trying to process an already-processed '
                'compilation database?'
            )

        clean_compdbs = CppCompilationDatabasesMap(settings)

        # Do processing, segregate processed commands into separate databases
        # for each target.
        for compile_command in self:
            processed_command = compile_command.process(
                default_path=default_path, path_globs=path_globs, strict=strict
            )

            if (
                processed_command is not None
                and processed_command.output_path is not None
            ):
                target = infer_target(
                    self.target_inference,
                    self._root_dir,
                    processed_command.output_path,
                )

                target = cast(str, target)
                processed_command.target = target
                clean_compdbs[target].add(processed_command)

                if clean_compdbs[target].source_file_path is None:
                    clean_compdbs[target].source_file_path = self.file_path
                    clean_compdbs[target].source_file_hash = self.file_hash

        # TODO(chadnorvell): Handle len(clean_compdbs) == 0

        # Determine if the processed database is functionally identical to the
        # original, unless configured to always output the new databases.
        # The criteria for "functionally identical" are:
        #
        # - The original file only contained commands for a single target
        # - The number of compile commands in the processed database is equal to
        #   that of the original database.
        #
        # This is a little bit crude. For example, it doesn't account for the
        # (rare) edge case of multiple databases having commands for the same
        # target. However, if you know that you have that kind of situation, you
        # should use `always_output_new` and not rely on this.
        if (
            not always_output_new
            and len(clean_compdbs) == 1
            and len(clean_compdbs[0]) == len(self)
        ):
            return None

        return clean_compdbs


class CppCompilationDatabasesMap:
    """Container for a map of target name to compilation database."""

    def __init__(self, settings: PigweedIdeSettings):
        self.settings = settings
        self._dbs: dict[str, CppCompilationDatabase] = dict()

    def __len__(self) -> int:
        return len(self._dbs)

    def _default(self, key: str | int):
        # This is like `defaultdict` except that we can use the provided key
        # (i.e. the target name) in the constructor.
        if isinstance(key, str) and key not in self._dbs:
            file_path = self.settings.working_dir / key / COMPDB_FILE_NAME
            self._dbs[key] = CppCompilationDatabase(file_path=file_path)

    def __getitem__(self, key: str | int) -> CppCompilationDatabase:
        self._default(key)

        # Support list-based indexing...
        if isinstance(key, int):
            return list(self._dbs.values())[key]

        # ... and key-based indexing.
        return self._dbs[key]

    def __setitem__(self, key: str, item: CppCompilationDatabase) -> None:
        self._default(key)
        self._dbs[key] = item

    def __iter__(self) -> Iterator[str]:
        for target, _ in self.items():
            yield target

    @property
    def targets(self) -> list[str]:
        return list(self._dbs.keys())

    def items(
        self,
    ) -> Generator[tuple[str, CppCompilationDatabase], None, None]:
        return ((key, value) for (key, value) in self._dbs.items())

    def _sort_by_commands(self) -> list[str]:
        """Sort targets by the number of compile commands they have."""
        enumerated_targets = sorted(
            [(len(db), target) for target, db in self._dbs.items()],
            key=lambda x: x[0],
            reverse=True,
        )

        return [target for (_, target) in enumerated_targets]

    def _sort_with_target_priority(self, target: str) -> list[str]:
        """Sorted targets, but with the provided target first."""
        sorted_targets = self._sort_by_commands()
        # This will raise a ValueError if the target is not in the list, but
        # we have ensured that that will never happen by the time we get here.
        sorted_targets.remove(target)
        return [target, *sorted_targets]

    def _targets_to_write(self, target: str) -> list[str]:
        """Return the list of targets whose comp. commands should be written.

        Under most conditions, this will return a list with just the provided
        target; essentially it's a no-op. But if ``cascade_targets`` is
        enabled, this returns a list of all targets with the provided target
        at the head of the list.
        """
        if not self.settings.cascade_targets:
            return [target]

        return self._sort_with_target_priority(target)

    def _compdb_to_write(self, target: str) -> CppCompilationDatabase:
        """The compilation database to write to file for this target.

        Under most conditions, this will return the compilation database
        associated with the provided target. But if ``cascade_targets`` is
        enabled, this returns a compilation database with commands from all
        targets, ordered per ``_sort_with_target_priority``.
        """
        targets = self._targets_to_write(target)
        compdb = CppCompilationDatabase()

        for iter_target in targets:
            compdb.add(*self[iter_target])

        return compdb

    def test_write(self) -> None:
        """Test writing to file.

        This will raise an exception if the file is not JSON-serializable."""
        for _, compdb in self.items():
            compdb.to_json()

    def write(self) -> None:
        """Write compilation databases to target-specific JSON files."""
        for target in self:
            path = self.settings.working_dir / target / COMPDB_FILE_NAME
            self._compdb_to_write(target).to_file(path)

    @classmethod
    def merge(
        cls, *db_sets: CppCompilationDatabasesMap
    ) -> CppCompilationDatabasesMap:
        """Merge several sets of processed compilation databases.

        If you process N compilation databases produced by a build system,
        you'll end up with N sets of processed compilation databases,
        containing databases for one or more targets each. This method
        merges them into one set of databases with one database per target.

        The expectation is that the vast majority of the time, each of the
        raw compilation databases that are processed will contain distinct
        targets, meaning that the keys of each ``CppCompilationDatabases``
        object that's merged will be unique to each object, and this operation
        is nothing more than a shallow merge.

        However, this also supports the case where targets may overlap between
        ``CppCompilationDatabases`` objects. In that case, we prioritize
        correctness, ensuring that the resulting compilation databases will
        work correctly with clangd. This means not including duplicate compile
        commands for the same file in the same target's database. The choice
        of which duplicate compile command ends up in the final database is
        unspecified and subject to change. Note also that this method expects
        the ``settings`` value to be the same between all of the provided
        ``CppCompilationDatabases`` objects.
        """
        if len(db_sets) == 0:
            raise ValueError(
                'At least one set of compilation databases is required.'
            )

        # Shortcut for the most common case.
        if len(db_sets) == 1:
            return db_sets[0]

        merged = cls(db_sets[0].settings)

        for dbs in db_sets:
            for target, db in dbs.items():
                merged[target].merge(db)

        return merged


@functools.lru_cache
def find_cipd_installed_exe_path(exe: str) -> Path:
    """Return the path of an executable installed by CIPD.

    Search for the executable in the paths pointed by all the defined
    `PW_<PROJ_NAME>_CIPD_INSTALL_DIR` environment variables.
    """

    if sys.platform.lower() in ("win32", "cygwin"):
        exe += ".exe"

    env_vars = vars(env)

    search_paths: list[str] = []
    for env_var_name, env_var in env_vars.items():
        if re.fullmatch(r"PW_[A-Z_]+_CIPD_INSTALL_DIR", env_var_name):
            search_paths.append(str(Path(env_var) / "bin" / exe))

    if (env_var := env_vars.get('PW_PIGWEED_CIPD_INSTALL_DIR')) is not None:
        search_paths.append(str(Path(env_var) / "bin" / exe))

    path = None
    exception = None
    try:
        path = path_to_executable(
            exe, default_path=None, path_globs=search_paths, strict=True
        )
    except UnresolvablePathException as e:
        exception = e

    if path is None or exception:
        search_paths_str = ":".join(search_paths)
        raise FileNotFoundError(
            f"Not able to find '{exe}' "
            f"among '{search_paths_str}'. Is bootstrap successful?"
        )

    return path


def get_clangd_path(settings: PigweedIdeSettings) -> Path:
    if settings.clangd_alternate_path is not None:
        return settings.clangd_alternate_path

    return find_cipd_installed_exe_path('clangd')


class ClangdSettings:
    """Makes system-specific settings for running ``clangd`` with Pigweed."""

    def __init__(self, settings: PigweedIdeSettings):
        state = CppIdeFeaturesState(settings)

        self.clangd_path = get_clangd_path(settings)

        compile_commands_dir = env.PW_PROJECT_ROOT

        if state.current_target is not None:
            compile_commands_dir = str(state.stable_target_link)

        host_cc_path = find_cipd_installed_exe_path("clang++")

        self.arguments: list[str] = [
            f'--compile-commands-dir={compile_commands_dir}',
            '--background-index',
            '--clang-tidy',
        ]

        query_driver = settings.clangd_query_driver_str(host_cc_path)

        if query_driver is not None:
            self.arguments.append(f'--query-driver={query_driver}')

    def command(self, system: str = platform.system()) -> str:
        """Return the command that runs clangd with Pigweed paths."""

        def make_command(line_continuation: str):
            arguments = f' {line_continuation}\n'.join(
                f'  {arg}' for arg in self.arguments
            )
            return f'\n{self.clangd_path} {line_continuation}\n{arguments}'

        if system.lower() == 'json':
            return '\n' + json.dumps(
                [str(self.clangd_path), *self.arguments], indent=2
            )

        if system.lower() in ['cmd', 'batch']:
            return make_command('`')

        if system.lower() in ['powershell', 'pwsh']:
            return make_command('^')

        if system.lower() == 'windows':
            return (
                f'\nIn PowerShell:\n{make_command("`")}'
                f'\n\nIn Command Prompt:\n{make_command("^")}'
            )

        # Default case for *sh-like shells.
        return make_command('\\')



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_ide settings."""

import enum
from inspect import cleandoc
import os
from pathlib import Path
from typing import Any, cast, Literal
import yaml

from pw_cli.env import pigweed_environment
from pw_config_loader.yaml_config_loader_mixin import YamlConfigLoaderMixin

env = pigweed_environment()
env_vars = vars(env)

PW_PROJECT_ROOT = Path(
    env.PW_PROJECT_ROOT if env.PW_PROJECT_ROOT is not None else os.getcwd()
)

PW_IDE_DIR_NAME = '.pw_ide'

_DEFAULT_BUILD_DIR_NAME = 'out'
_DEFAULT_BUILD_DIR = PW_PROJECT_ROOT / _DEFAULT_BUILD_DIR_NAME

_DEFAULT_WORKSPACE_ROOT = PW_PROJECT_ROOT

_DEFAULT_TARGET_INFERENCE = '?'

SupportedEditorName = Literal['vscode']


class SupportedEditor(enum.Enum):
    VSCODE = 'vscode'


_DEFAULT_SUPPORTED_EDITORS: dict[SupportedEditorName, bool] = {
    'vscode': True,
}

_DEFAULT_CONFIG: dict[str, Any] = {
    'cascade_targets': False,
    'clangd_alternate_path': None,
    'clangd_additional_query_drivers': [],
    'compdb_gen_cmd': None,
    'compdb_search_paths': [_DEFAULT_BUILD_DIR_NAME],
    'default_target': None,
    'workspace_root': _DEFAULT_WORKSPACE_ROOT,
    'editors': _DEFAULT_SUPPORTED_EDITORS,
    'sync': ['pw --no-banner ide cpp --process'],
    'targets_exclude': [],
    'targets_include': [],
    'target_inference': _DEFAULT_TARGET_INFERENCE,
    'working_dir': _DEFAULT_WORKSPACE_ROOT / PW_IDE_DIR_NAME,
}

_DEFAULT_PROJECT_FILE = PW_PROJECT_ROOT / '.pw_ide.yaml'
_DEFAULT_PROJECT_USER_FILE = PW_PROJECT_ROOT / '.pw_ide.user.yaml'
_DEFAULT_USER_FILE = Path.home() / '.pw_ide.yaml'


def _expand_any_vars(input_path: Path) -> Path:
    """Expand any environment variables in a path.

    Python's ``os.path.expandvars`` will only work on an isolated environment
    variable name. In shell, you can expand variables within a larger command
    or path. We replicate that functionality here.
    """
    outputs = []

    for token in input_path.parts:
        expanded_var = os.path.expandvars(token)

        if expanded_var == token:
            outputs.append(token)
        else:
            outputs.append(expanded_var)

    # pylint: disable=no-value-for-parameter
    return Path(os.path.join(*outputs))
    # pylint: enable=no-value-for-parameter


def _expand_any_vars_str(input_path: str) -> str:
    """`_expand_any_vars`, except takes and returns a string instead of path."""
    return str(_expand_any_vars(Path(input_path)))


def _parse_dir_path(input_path_str: str, workspace_root: Path) -> Path:
    if (path := Path(input_path_str)).is_absolute():
        return path

    return (workspace_root / path).resolve()


def _parse_compdb_search_path(
    input_data: str | tuple[str, str],
    default_inference: str,
    workspace_root: Path,
) -> tuple[Path, str]:
    if isinstance(input_data, (tuple, list)):
        return _parse_dir_path(input_data[0], workspace_root), input_data[1]

    return _parse_dir_path(input_data, workspace_root), default_inference


class PigweedIdeSettings(YamlConfigLoaderMixin):
    """Pigweed IDE features settings storage class."""

    def __init__(
        self,
        project_file: Path | bool = _DEFAULT_PROJECT_FILE,
        project_user_file: Path | bool = _DEFAULT_PROJECT_USER_FILE,
        user_file: Path | bool = _DEFAULT_USER_FILE,
        default_config: dict[str, Any] | None = None,
    ) -> None:
        self.config_init(
            config_section_title='pw_ide',
            project_file=project_file,
            project_user_file=project_user_file,
            user_file=user_file,
            default_config=_DEFAULT_CONFIG
            if default_config is None
            else default_config,
            environment_var='PW_IDE_CONFIG_FILE',
        )

        # YamlConfigLoaderMixin only conditionally assigns this, but we rely
        # on it later, so we just extra assign it here.
        self.project_file = project_file  # type: ignore

    def __repr__(self) -> str:
        return str(
            {
                key: getattr(self, key)
                for key, value in self.__class__.__dict__.items()
                if isinstance(value, property)
            }
        )

    @property
    def working_dir(self) -> Path:
        """Path to the ``pw_ide`` working directory.

        The working directory holds C++ compilation databases and caches, and
        other supporting files. This should not be a directory that's regularly
        deleted or manipulated by other processes (e.g. the GN ``out``
        directory) nor should it be committed to source control.
        """
        return Path(
            _expand_any_vars_str(
                self._config.get(
                    'working_dir', self.workspace_root / PW_IDE_DIR_NAME
                )
            )
        )

    @property
    def compdb_gen_cmd(self) -> str | None:
        """The command that should be run to generate a compilation database.

        Defining this allows ``pw_ide`` to automatically generate a compilation
        database if it suspects one has not been generated yet before a sync.
        """
        return self._config.get('compdb_gen_cmd')

    @property
    def compdb_search_paths(self) -> list[tuple[Path, str]]:
        """Paths to directories to search for compilation databases.

        If you're using a build system to generate compilation databases, this
        may simply be your build output directory. However, you can add
        additional directories to accommodate compilation databases from other
        sources.

        Entries can be just directories, in which case the default target
        inference pattern will be used. Or entries can be tuples of a directory
        and a target inference pattern. See the documentation for
        ``target_inference`` for more information.

        Finally, the directories can be concrete paths, or they can be globs
        that expand to multiple paths.

        Note that relative directory paths will be resolved relative to the
        workspace root.
        """
        return [
            _parse_compdb_search_path(
                search_path, self.target_inference, self.workspace_root
            )
            for search_path in self._config.get(
                'compdb_search_paths', [_DEFAULT_BUILD_DIR]
            )
        ]

    @property
    def targets_exclude(self) -> list[str]:
        """The list of targets that should not be enabled for code analysis.

        In this case, "target" is analogous to a GN target, i.e., a particular
        build configuration. By default, all available targets are enabled. By
        adding targets to this list, you can disable/hide targets that should
        not be available for code analysis.

        Target names need to match the name of the directory that holds the
        build system artifacts for the target. For example, GN outputs build
        artifacts for the ``pw_strict_host_clang_debug`` target in a directory
        with that name in its output directory. So that becomes the canonical
        name for the target.
        """
        return self._config.get('targets_exclude', list())

    @property
    def targets_include(self) -> list[str]:
        """The list of targets that should be enabled for code analysis.

        In this case, "target" is analogous to a GN target, i.e., a particular
        build configuration. By default, all available targets are enabled. By
        adding targets to this list, you can constrain the targets that are
        enabled for code analysis to a subset of those that are available, which
        may be useful if your project has many similar targets that are
        redundant from a code analysis perspective.

        Target names need to match the name of the directory that holds the
        build system artifacts for the target. For example, GN outputs build
        artifacts for the ``pw_strict_host_clang_debug`` target in a directory
        with that name in its output directory. So that becomes the canonical
        name for the target.
        """
        return self._config.get('targets_include', list())

    @property
    def target_inference(self) -> str:
        """A glob-like string for extracting a target name from an output path.

        Build systems and projects have varying ways of organizing their build
        directory structure. For a given compilation unit, we need to know how
        to extract the build's target name from the build artifact path. A
        simple example:

        .. code-block:: none

           clang++ hello.cc -o host/obj/hello.cc.o

        The top-level directory ``host`` is the target name we want. The same
        compilation unit might be used with another build target:

        .. code-block:: none

           gcc-arm-none-eabi hello.cc -o arm_dev_board/obj/hello.cc.o

        In this case, this compile command is associated with the
        ``arm_dev_board`` target.

        When importing and processing a compilation database, we assume by
        default that for each compile command, the corresponding target name is
        the name of the top level directory within the build directory root
        that contains the build artifact. This is the default behavior for most
        build systems. However, if your project is structured differently, you
        can provide a glob-like string that indicates how to extract the target
        name from build artifact path.

        A ``*`` indicates any directory, and ``?`` indicates the directory that
        has the name of the target. The path is resolved from the build
        directory root, and anything deeper than the target directory is
        ignored. For example, a glob indicating that the directory two levels
        down from the build directory root has the target name would be
        expressed with ``*/*/?``.

        Note that the build artifact path is relative to the compilation
        database search path that found the file. For example, for a compilation
        database search path of ``{project dir}/out``, for the purposes of
        target inference, the build artifact path is relative to the ``{project
        dir}/out`` directory. Target inference patterns can be defined for each
        compilation database search path. See the documentation for
        ``compdb_search_paths`` for more information.
        """
        return self._config.get('target_inference', _DEFAULT_TARGET_INFERENCE)

    @property
    def default_target(self) -> str | None:
        """The default target to use when calling ``--set-default``.

        This target will be selected when ``pw ide cpp --set-default`` is
        called. You can define an explicit default target here. If that command
        is invoked without a default target definition, ``pw_ide`` will try to
        infer the best choice of default target. Currently, it selects the
        target with the broadest compilation unit coverage.
        """
        return self._config.get('default_target', None)

    @property
    def sync(self) -> list[str]:
        """A sequence of commands to automate IDE features setup.

        ``pw ide sync`` should do everything necessary to get the project from
        a fresh checkout to a working default IDE experience. This defines the
        list of commands that makes that happen, which will be executed
        sequentially in subprocesses. These commands should be idempotent, so
        that the user can run them at any time to update their IDE features
        configuration without the risk of putting those features in a bad or
        unexpected state.
        """
        return self._config.get('sync', list())

    @property
    def clangd_alternate_path(self) -> Path | None:
        """An alternate path to ``clangd`` to use instead of Pigweed's.

        Pigweed provides the ``clang`` toolchain, including ``clangd``, via
        CIPD, and by default, ``pw_ide`` will look for that toolchain in the
        CIPD directory at ``$PW_PIGWEED_CIPD_INSTALL_DIR`` *or* in an alternate
        CIPD directory specified by ``$PW_{project name}_CIPD_INSTALL_DIR`` if
        it exists.

        If your project needs to use a ``clangd`` located somewhere else not
        covered by the cases described above, you can define the path to that
        ``clangd`` here.
        """
        return self._config.get('clangd_alternate_path', None)

    @property
    def clangd_additional_query_drivers(self) -> list[str] | None:
        """Additional query driver paths that clangd should use.

        By default, ``pw_ide`` supplies driver paths for the toolchains included
        in Pigweed. If you are using toolchains that are not supplied by
        Pigweed, you should include path globs to your toolchains here. These
        paths will be given higher priority than the Pigweed toolchain paths.

        If you want to omit the query drivers argument altogether, set this to
        ``null``.
        """
        return self._config.get('clangd_additional_query_drivers', list())

    def clangd_query_drivers(
        self, host_clang_cc_path: Path
    ) -> list[str] | None:
        if self.clangd_additional_query_drivers is None:
            return None

        drivers = [
            *[
                _expand_any_vars_str(p)
                for p in self.clangd_additional_query_drivers
            ],
        ]

        drivers.append(str(host_clang_cc_path.parent / '*'))

        if (env_var := env_vars.get('PW_ARM_CIPD_INSTALL_DIR')) is not None:
            drivers.append(str(Path(env_var) / 'bin' / '*'))

        return drivers

    def clangd_query_driver_str(self, host_clang_cc_path: Path) -> str | None:
        clangd_query_drivers = self.clangd_query_drivers(host_clang_cc_path)

        if clangd_query_drivers is None:
            return None

        return ','.join(clangd_query_drivers)

    @property
    def workspace_root(self) -> Path:
        """The root directory of the IDE workspace.

        In most cases, the IDE workspace directory is also your Pigweed project
        root; that's the default, and if that's the case, this you can omit this
        configuration.

        If your project has a structure where the directory you open in your
        IDE is *not* the Pigweed project root directory, you can specify the
        workspace root directory here to ensure that IDE support files are
        put in the right place.

        For example, given this directory structure:

        .. code-block::

           my_project
           |
           ├- third_party
           ├- docs
           ├- src
           |  ├- pigweed.json
           |  └- ... all other project source files
           |
           └- pigweed
               └ ... upstream Pigweed source, e.g. a submodule

        In this case ```my_project/src/``` is the Pigweed project root, but
        ```my_project/``` is the workspace root directory you will open in your
        IDE.

        A relative path will be resolved relative to the directory in which the
        ``.pw_ide.yaml`` config file is located.
        """
        workspace_root = Path(
            self._config.get('workspace_root', _DEFAULT_WORKSPACE_ROOT)
        )

        project_file_exists = (
            isinstance(self.project_file, Path) and self.project_file.exists()
        )

        config_file_root = (
            cast(Path, self.project_file)
            if project_file_exists
            else _DEFAULT_PROJECT_FILE
        ).parent

        if workspace_root.is_absolute():
            return workspace_root

        return config_file_root.resolve() / workspace_root

    @property
    def editors(self) -> dict[str, bool]:
        """Enable or disable automated support for editors.

        Automatic support for some editors is provided by ``pw_ide``, which is
        accomplished through generating configuration files in your project
        directory. All supported editors are enabled by default, but you can
        disable editors by adding an ``'<editor>': false`` entry.
        """
        return self._config.get('editors', _DEFAULT_SUPPORTED_EDITORS)

    def editor_enabled(self, editor: SupportedEditorName) -> bool:
        """True if the provided editor is enabled in settings.

        This module will integrate the project with all supported editors by
        default. If the project or user want to disable particular editors,
        they can do so in the appropriate settings file.
        """
        return self._config.get('editors', {}).get(editor, False)

    @property
    def cascade_targets(self) -> bool:
        """Mix compile commands for multiple targets to maximize code coverage.

        By default (with this set to ``False``), the compilation database for
        each target is consistent in the sense that it only contains compile
        commands for one build target, so the code intelligence that database
        provides is related to a single, known compilation artifact. However,
        this means that code intelligence may not be provided for every source
        file in a project, because some source files may be relevant to targets
        other than the one you have currently set. Those source files won't
        have compile commands for the current target, and no code intelligence
        will appear in your editor.

        If this is set to ``True``, compilation databases will still be
        separated by target, but compile commands for *all other targets* will
        be appended to the list of compile commands for *that* target. This
        will maximize code coverage, ensuring that you have code intelligence
        for every file that is built for any target, at the cost of
        consistency—the code intelligence for some files may show information
        that is incorrect or irrelevant to the currently selected build target.

        The currently set target's compile commands will take priority at the
        top of the combined file, then all other targets' commands will come
        after in order of the number of commands they have (i.e. in the order of
        their code coverage). This relies on the fact that ``clangd`` parses the
        compilation database from the top down, using the first compile command
        it encounters for each compilation unit.
        """
        return self._config.get('cascade_targets', False)


def _docstring_set_default(
    obj: Any, default: Any, literal: bool = False
) -> None:
    """Add a default value annotation to a docstring.

    Formatting isn't allowed in docstrings, so by default we can't inject
    variables that we would like to appear in the documentation, like the
    default value of a property. But we can use this function to add it
    separately.
    """
    if obj.__doc__ is not None:
        default = str(default)

        if literal:
            lines = default.splitlines()

            if len(lines) == 0:
                return
            if len(lines) == 1:
                default = f'Default: ``{lines[0]}``'
            else:
                default = 'Default:\n\n.. code-block::\n\n  ' + '\n  '.join(
                    lines
                )

        doc = cast(str, obj.__doc__)
        obj.__doc__ = f'{cleandoc(doc)}\n\n{default}'


_docstring_set_default(
    PigweedIdeSettings.working_dir, PW_IDE_DIR_NAME, literal=True
)
_docstring_set_default(
    PigweedIdeSettings.compdb_gen_cmd,
    _DEFAULT_CONFIG['compdb_gen_cmd'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.compdb_search_paths,
    [_DEFAULT_BUILD_DIR_NAME],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.targets_exclude,
    _DEFAULT_CONFIG['targets_exclude'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.targets_include,
    _DEFAULT_CONFIG['targets_include'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.default_target,
    _DEFAULT_CONFIG['default_target'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.cascade_targets,
    _DEFAULT_CONFIG['cascade_targets'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.target_inference,
    _DEFAULT_CONFIG['target_inference'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.sync, _DEFAULT_CONFIG['sync'], literal=True
)
_docstring_set_default(
    PigweedIdeSettings.clangd_alternate_path,
    _DEFAULT_CONFIG['clangd_alternate_path'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.clangd_additional_query_drivers,
    _DEFAULT_CONFIG['clangd_additional_query_drivers'],
    literal=True,
)
_docstring_set_default(
    PigweedIdeSettings.editors,
    yaml.dump(_DEFAULT_SUPPORTED_EDITORS),
    literal=True,
)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Code formatter plugin for Bazel build files."""

from pathlib import Path
from typing import Dict, Final, Iterable, Iterator, List, Sequence, Tuple

from pw_presubmit.format.core import (
    FileFormatter,
    FormattedFileContents,
    FormatFixStatus,
)


class BuildifierFormatter(FileFormatter):
    """A formatter that runs ``buildifier`` on files."""

    # These warnings are safe to enable because they can always be auto-fixed.
    DEFAULT_WARNINGS_TO_FIX: Final[Sequence[str]] = (
        'load',
        'load-on-top',
        'native-build',
        'same-origin-load',
        'out-of-order-load',
        'unsorted-dict-items',
    )

    def __init__(
        self, warnings_to_fix: Sequence[str] = DEFAULT_WARNINGS_TO_FIX, **kwargs
    ):
        super().__init__(**kwargs)
        self.warnings_to_fix = list(warnings_to_fix)

    @staticmethod
    def _detect_file_type(file_path: Path) -> str:
        if file_path.name == 'MODULE.bazel':
            return 'module'
        if file_path.name == 'WORKSPACE':
            return 'workspace'
        if '.bzl' in file_path.name:
            return 'bzl'
        if 'BUILD' in file_path.name or file_path.suffix == '.bazel':
            return 'build'

        return 'default'

    def _files_by_type(self, paths: Iterable[Path]) -> Dict[str, List[Path]]:
        all_types = (
            'module',
            'workspace',
            'bzl',
            'build',
            'default',
        )
        all_files: Dict[str, List[Path]] = {t: [] for t in all_types}

        for file_path in paths:
            all_files[self._detect_file_type(file_path)].append(file_path)

        return all_files

    def format_file_in_memory(
        self, file_path: Path, file_contents: bytes
    ) -> FormattedFileContents:
        """Uses ``buildifier`` to check the formatting of the requested file.

        The file at ``file_path`` is NOT modified by this check.

        Returns:
            A populated
            :py:class:`pw_presubmit.format.core.FormattedFileContents` that
            contains either the result of formatting the file, or an error
            message.
        """
        proc = self.run_tool(
            'buildifier',
            [
                f'--type={self._detect_file_type(file_path)}',
                '--lint=fix',
                '--warnings=' + ','.join(self.warnings_to_fix),
            ],
            input=file_contents,
        )
        ok = proc.returncode == 0
        return FormattedFileContents(
            ok=ok,
            formatted_file_contents=proc.stdout,
            error_message=None if ok else proc.stderr.decode(),
        )

    def format_file(self, file_path: Path) -> FormatFixStatus:
        """Formats the provided file in-place using ``buildifier``.

        Returns:
            A FormatFixStatus that contains relevant errors/warnings.
        """
        proc = self.run_tool(
            'buildifier',
            [
                f'--type={self._detect_file_type(file_path)}',
                '--lint=fix',
                '--warnings=' + ','.join(self.warnings_to_fix),
                file_path,
            ],
        )
        ok = proc.returncode == 0
        return FormatFixStatus(
            ok=ok,
            error_message=None if ok else proc.stderr.decode(),
        )

    def format_files(
        self, paths: Iterable[Path], keep_warnings: bool = True
    ) -> Iterator[Tuple[Path, FormatFixStatus]]:
        """Uses ``buildifier`` to format the specified files in-place.

        Returns:
            An iterator of ``Path`` and
            :py:class:`pw_presubmit.format.core.FormatFixStatus` pairs for each
            file that was not successfully formatted. If ``keep_warnings`` is
            ``True``, any successful format operations with warnings will also
            be returned.
        """
        sorted_files = self._files_by_type(paths)
        for file_type, type_specific_paths in sorted_files.items():
            if not type_specific_paths:
                continue

            proc = self.run_tool(
                'buildifier',
                [
                    f'--type={file_type}',
                    '--lint=fix',
                    '--warnings=' + ','.join(self.warnings_to_fix),
                    *type_specific_paths,
                ],
            )

            # If there's an error, fall back to per-file formatting to figure
            # out which file has problems.
            if proc.returncode != 0:
                yield from super().format_files(
                    type_specific_paths, keep_warnings
                )



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines arguments for the pw command."""

import argparse
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
import logging
from pathlib import Path
import sys
from typing import Any, NoReturn

from pw_cli import argument_types, plugins
from pw_cli.branding import banner
import pw_cli.env

_HELP_HEADER = '''The Pigweed command line interface (CLI).

Example uses:
    pw logdemo
    pw --loglevel debug watch -C out
'''


def parse_args() -> argparse.Namespace:
    return arg_parser().parse_args()


class ShellCompletionFormat(Enum):
    """Supported shell tab completion modes."""

    BASH = 'bash'
    FISH = 'fish'
    ZSH = 'zsh'


@dataclass(frozen=True)
class ShellCompletion:
    """Transforms argparse actions into bash, fish, zsh shell completions."""

    action: argparse.Action
    parser: argparse.ArgumentParser

    @property
    def option_strings(self) -> list[str]:
        return list(self.action.option_strings)

    @cached_property
    def help(self) -> str:
        return self.parser._get_formatter()._expand_help(  # pylint: disable=protected-access
            self.action
        )

    @property
    def choices(self) -> list[str]:
        return list(self.action.choices) if self.action.choices else []

    @property
    def flag(self) -> bool:
        return self.action.nargs == 0

    @property
    def default(self) -> Any:
        return self.action.default

    def bash_option(self, text: str) -> list[str]:
        result: list[str] = []
        for option_str in self.option_strings:
            if option_str.startswith(text):
                result.append(option_str)
        return result

    def zsh_option(self, text: str) -> list[str]:
        result: list[str] = []
        for option_str in self.option_strings:
            if not option_str.startswith(text):
                continue

            short_and_long_opts = ' '.join(self.option_strings)
            # '(-h --help)-h[Display help message and exit]'
            # '(-h --help)--help[Display help message and exit]'
            help_text = self.help if self.help else ''

            state_str = ''
            if not self.flag:
                state_str = ': :->' + option_str

            result.append(
                f'({short_and_long_opts}){option_str}[{help_text}]'
                f'{state_str}'
            )
        return result

    def fish_option(self, text: str) -> list[str]:
        result: list[str] = []
        for option_str in self.option_strings:
            if not option_str.startswith(text):
                continue

            output: list[str] = []
            if option_str.startswith('--'):
                output.append(f'--long-option\t{option_str.lstrip("-")}')
            elif option_str.startswith('-'):
                output.append(f'--short-option\t{option_str.lstrip("-")}')

            if self.choices:
                choice_str = " ".join(self.choices)
                output.append('--exclusive')
                output.append('--arguments')
                output.append(f'(string split " " "{choice_str}")')
            elif self.action.type == Path:
                output.append('--require-parameter')
                output.append('--force-files')

            if self.help:
                output.append(f'--description\t"{self.help}"')

            result.append('\t'.join(output))
        return result


def get_options_and_help(
    parser: argparse.ArgumentParser,
) -> list[ShellCompletion]:
    return list(
        ShellCompletion(action=action, parser=parser)
        for action in parser._actions  # pylint: disable=protected-access
    )


def print_completions_for_option(
    parser: argparse.ArgumentParser,
    text: str = '',
    tab_completion_format: str = ShellCompletionFormat.BASH.value,
) -> None:
    matched_lines: list[str] = []
    for completion in get_options_and_help(parser):
        if tab_completion_format == ShellCompletionFormat.ZSH.value:
            matched_lines.extend(completion.zsh_option(text))
        if tab_completion_format == ShellCompletionFormat.FISH.value:
            matched_lines.extend(completion.fish_option(text))
        else:
            matched_lines.extend(completion.bash_option(text))

    for line in matched_lines:
        print(line)


def print_banner() -> None:
    """Prints the PIGWEED (or project specific) banner to stderr."""
    parsed_env = pw_cli.env.pigweed_environment()
    if parsed_env.PW_ENVSETUP_NO_BANNER or parsed_env.PW_ENVSETUP_QUIET:
        return
    print(banner() + '\n', file=sys.stderr)


def format_help(registry: plugins.Registry) -> str:
    """Returns the pw help information as a string."""
    return f'{arg_parser().format_help()}\n{registry.short_help()}'


class _ArgumentParserWithBanner(argparse.ArgumentParser):
    """Parser that the Pigweed banner when there are parsing errors."""

    def error(self, message: str) -> NoReturn:
        print_banner()
        self.print_usage(sys.stderr)
        self.exit(2, f'{self.prog}: error: {message}\n')


def add_tab_complete_arguments(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        '--tab-complete-option',
        nargs='?',
        help='Print tab completions for the supplied option text.',
    )
    parser.add_argument(
        '--tab-complete-format',
        choices=list(shell.value for shell in ShellCompletionFormat),
        default='bash',
        help='Output format for tab completion results.',
    )
    return parser


def arg_parser() -> argparse.ArgumentParser:
    """Creates an argument parser for the pw command."""
    argparser = _ArgumentParserWithBanner(
        prog='pw',
        add_help=False,
        description=_HELP_HEADER,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Do not use the built-in help argument so that displaying the help info can
    # be deferred until the pw plugins have been registered.
    argparser.add_argument(
        '-h',
        '--help',
        action='store_true',
        help='Display this help message and exit',
    )
    argparser.add_argument(
        '-C',
        '--directory',
        type=argument_types.directory,
        default=Path.cwd(),
        help='Change to this directory before doing anything',
    )
    argparser.add_argument(
        '-l',
        '--loglevel',
        type=argument_types.log_level,
        default=logging.INFO,
        help='Set the log level (debug, info, warning, error, critical)',
    )
    argparser.add_argument(
        '--debug-log',
        type=Path,
        help=(
            'Additionally log to this file at debug level; does not affect '
            'terminal output'
        ),
    )
    argparser.add_argument(
        '--banner',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Whether to print the Pigweed banner',
    )
    argparser.add_argument(
        '--tab-complete-command',
        nargs='?',
        help='Print tab completions for the supplied command text.',
    )
    argparser.add_argument(
        'command',
        nargs='?',
        help='Which command to run; see supported commands below',
    )
    argparser.add_argument(
        'plugin_args',
        metavar='...',
        nargs=argparse.REMAINDER,
        help='Remaining arguments are forwarded to the command',
    )
    return add_tab_complete_arguments(argparser)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates the modules index table on //docs/modules.rst."""

import json
import os
import sys

from sphinx.application import Sphinx


with open(f'{os.environ["PW_ROOT"]}/PIGWEED_MODULES', 'r') as f:
    # The complete, authoritative list of modules.
    complete_pigweed_modules_list = f.read().splitlines()
with open(f'{os.environ["PW_ROOT"]}/docs/module_metadata.json', 'r') as f:
    # Module metadata such as supported languages and status.
    metadata = json.load(f)


def build_status_badge(status):
    """Styles the module status as a clickable badge."""
    # This default value should never get used but it's a valid
    # neutral styling in case something goes wrong and it leaks through.
    badge_type = 'info'
    if status == 'stable':
        badge_type = 'primary'
    elif status == 'unstable':
        badge_type = 'secondary'
    elif status == 'experimental':
        badge_type = 'warning'
    elif status == 'deprecated':
        badge_type = 'danger'
    else:
        msg = f'[modules_index.py] error: invalid module status ("{status}")'
        sys.exit(msg)
    # Use bdg-ref-* to make the badge link to the glossary definition for
    # "stable", "experimental", etc.
    # https://sphinx-design.readthedocs.io/en/latest/badges_buttons.html
    return f':bdg-{badge_type}:`{status}`'


def build_row(module_name: str):
    """Builds a row of data for the table. Each module gets a row."""
    ref = f':ref:`module-{module_name}`'
    if module_name not in metadata:
        return f'   "{ref}", "", "", ""\n'
    tagline = metadata[module_name]['tagline']
    status = build_status_badge(metadata[module_name]['status'])
    if 'languages' in metadata[module_name]:
        languages = ', '.join(metadata[module_name]['languages'])
    else:
        languages = ''
    return f'   "{ref}", "{status}", "{tagline}", "{languages}"\n'


def generate_modules_index(_, docname: str, source: list[str]) -> None:
    """Inserts the metadata table into //docs/modules.rst."""
    if docname != 'modules':  # Only run this logic on //docs/modules.rst.
        return
    skip = ['docker']  # Items in //PIGWEED_MODULES that should be skipped.
    # Transform //docs/module_metadata.json into a csv-table reST directive.
    # https://docutils.sourceforge.io/docs/ref/rst/directives.html#csv-table-1
    content = '\n\n.. csv-table::\n'
    content += '   :header: "Name", "Status", "Description", "Languages"\n\n'
    # Loop through the complete, authoritative list (as opposed to the metadata)
    # to guarantee that every module is listed on the modules index page.
    for module in complete_pigweed_modules_list:
        if module in skip:
            continue
        content += build_row(module)
    # Modify the reST of //docs/modules.rst in-place. The auto-generated table
    # is just appended to the reST source text.
    source[0] += content


def setup(app: Sphinx) -> dict[str, bool]:
    """Hooks this extension into Sphinx."""
    app.connect('source-read', generate_modules_index)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_presubmit ContextVar."""

from __future__ import annotations

from contextvars import ContextVar
import dataclasses
import enum
import inspect
import logging
import json
import os
from pathlib import Path
import re
import shlex
import shutil
import subprocess
import tempfile
from typing import (
    Any,
    Iterable,
    NamedTuple,
    Sequence,
    TYPE_CHECKING,
)
import urllib

import pw_cli.color
import pw_cli.env
import pw_env_setup.config_file

if TYPE_CHECKING:
    from pw_presubmit.presubmit import Check

_COLOR = pw_cli.color.colors()
_LOG: logging.Logger = logging.getLogger(__name__)

PRESUBMIT_CHECK_TRACE: ContextVar[
    dict[str, list[PresubmitCheckTrace]]
] = ContextVar('pw_presubmit_check_trace', default={})


@dataclasses.dataclass(frozen=True)
class FormatOptions:
    python_formatter: str | None = 'black'
    black_path: str | None = 'black'
    exclude: Sequence[re.Pattern] = dataclasses.field(default_factory=list)

    @staticmethod
    def load(env: dict[str, str] | None = None) -> FormatOptions:
        config = pw_env_setup.config_file.load(env=env)
        fmt = config.get('pw', {}).get('pw_presubmit', {}).get('format', {})
        return FormatOptions(
            python_formatter=fmt.get('python_formatter', 'black'),
            black_path=fmt.get('black_path', 'black'),
            exclude=tuple(re.compile(x) for x in fmt.get('exclude', ())),
        )

    def filter_paths(self, paths: Iterable[Path]) -> tuple[Path, ...]:
        root = Path(pw_cli.env.pigweed_environment().PW_PROJECT_ROOT)
        relpaths = [x.relative_to(root) for x in paths]

        for filt in self.exclude:
            relpaths = [x for x in relpaths if not filt.search(str(x))]
        return tuple(root / x for x in relpaths)


def get_buildbucket_info(bbid) -> dict[str, Any]:
    if not bbid or not shutil.which('bb'):
        return {}

    output = subprocess.check_output(
        ['bb', 'get', '-json', '-p', f'{bbid}'], text=True
    )
    return json.loads(output)


@dataclasses.dataclass
class LuciPipeline:
    """Details of previous builds in this pipeline, if applicable.

    Attributes:
        round: The zero-indexed round number.
        builds_from_previous_iteration: A list of the buildbucket ids from the
            previous round, if any.
    """

    round: int
    builds_from_previous_iteration: Sequence[int]

    @staticmethod
    def create(
        bbid: int,
        fake_pipeline_props: dict[str, Any] | None = None,
    ) -> LuciPipeline | None:
        pipeline_props: dict[str, Any]
        if fake_pipeline_props is not None:
            pipeline_props = fake_pipeline_props
        else:
            pipeline_props = (
                get_buildbucket_info(bbid)
                .get('input', {})
                .get('properties', {})
                .get('$pigweed/pipeline', {})
            )
        if not pipeline_props.get('inside_a_pipeline', False):
            return None

        return LuciPipeline(
            round=int(pipeline_props['round']),
            builds_from_previous_iteration=list(
                int(x) for x in pipeline_props['builds_from_previous_iteration']
            ),
        )


@dataclasses.dataclass
class LuciTrigger:
    """Details the pending change or submitted commit triggering the build.

    Attributes:
        number: The number of the change in Gerrit.
        patchset: The number of the patchset of the change.
        remote: The full URL of the remote.
        project: The name of the project in Gerrit.
        branch: The name of the branch on which this change is being/was
            submitted.
        ref: The "refs/changes/.." path that can be used to reference the
            patch for unsubmitted changes and the hash for submitted changes.
        gerrit_name: The name of the googlesource.com Gerrit host.
        submitted: Whether the change has been submitted or is still pending.
        gerrit_host: The scheme and hostname of the googlesource.com Gerrit
            host.
        gerrit_url: The full URL to this change on the Gerrit host.
        gitiles_url: The full URL to this commit in Gitiles.
    """

    number: int
    patchset: int
    remote: str
    project: str
    branch: str
    ref: str
    gerrit_name: str
    submitted: bool

    @property
    def gerrit_host(self):
        return f'https://{self.gerrit_name}-review.googlesource.com'

    @property
    def gerrit_url(self):
        if not self.number:
            return self.gitiles_url
        return f'{self.gerrit_host}/c/{self.number}'

    @property
    def gitiles_url(self):
        return f'{self.remote}/+/{self.ref}'

    @staticmethod
    def create_from_environment(
        env: dict[str, str] | None = None,
    ) -> Sequence['LuciTrigger']:
        if not env:
            env = os.environ.copy()
        raw_path = env.get('TRIGGERING_CHANGES_JSON')
        if not raw_path:
            return ()
        path = Path(raw_path)
        if not path.is_file():
            return ()

        result = []
        with open(path, 'r') as ins:
            for trigger in json.load(ins):
                keys = {
                    'number',
                    'patchset',
                    'remote',
                    'project',
                    'branch',
                    'ref',
                    'gerrit_name',
                    'submitted',
                }
                if keys <= trigger.keys():
                    result.append(LuciTrigger(**{x: trigger[x] for x in keys}))

        return tuple(result)

    @staticmethod
    def create_for_testing(**kwargs):
        change = {
            'number': 123456,
            'patchset': 1,
            'remote': 'https://pigweed.googlesource.com/pigweed/pigweed',
            'project': 'pigweed/pigweed',
            'branch': 'main',
            'ref': 'refs/changes/56/123456/1',
            'gerrit_name': 'pigweed',
            'submitted': True,
        }
        change.update(kwargs)

        with tempfile.TemporaryDirectory() as tempdir:
            changes_json = Path(tempdir) / 'changes.json'
            with changes_json.open('w') as outs:
                json.dump([change], outs)
            env = {'TRIGGERING_CHANGES_JSON': changes_json}
            return LuciTrigger.create_from_environment(env)


@dataclasses.dataclass
class LuciContext:
    """LUCI-specific information about the environment.

    Attributes:
        buildbucket_id: The globally-unique buildbucket id of the build.
        build_number: The builder-specific incrementing build number, if
            configured for this builder.
        project: The LUCI project under which this build is running (often
            "pigweed" or "pigweed-internal").
        bucket: The LUCI bucket under which this build is running (often ends
            with "ci" or "try").
        builder: The builder being run.
        swarming_server: The swarming server on which this build is running.
        swarming_task_id: The swarming task id of this build.
        cas_instance: The CAS instance accessible from this build.
        context_file: The path to the LUCI_CONTEXT file.
        pipeline: Information about the build pipeline, if applicable.
        triggers: Information about triggering commits, if applicable.
        is_try: True if the bucket is a try bucket.
        is_ci: True if the bucket is a ci bucket.
        is_dev: True if the bucket is a dev bucket.
        is_shadow: True if the bucket is a shadow bucket.
        is_prod: True if both is_dev and is_shadow are False.
    """

    buildbucket_id: int
    build_number: int
    project: str
    bucket: str
    builder: str
    swarming_server: str
    swarming_task_id: str
    cas_instance: str
    context_file: Path
    pipeline: LuciPipeline | None
    triggers: Sequence[LuciTrigger] = dataclasses.field(default_factory=tuple)

    @property
    def is_try(self):
        return re.search(r'\btry$', self.bucket)

    @property
    def is_ci(self):
        return re.search(r'\bci$', self.bucket)

    @property
    def is_dev(self):
        return re.search(r'\bdev\b', self.bucket)

    @property
    def is_shadow(self):
        return re.search(r'\bshadow\b', self.bucket)

    @property
    def is_prod(self):
        return not self.is_dev and not self.is_shadow

    @staticmethod
    def create_from_environment(
        env: dict[str, str] | None = None,
        fake_pipeline_props: dict[str, Any] | None = None,
    ) -> LuciContext | None:
        """Create a LuciContext from the environment."""

        if not env:
            env = os.environ.copy()

        luci_vars = [
            'BUILDBUCKET_ID',
            'BUILDBUCKET_NAME',
            'BUILD_NUMBER',
            'LUCI_CONTEXT',
            'SWARMING_TASK_ID',
            'SWARMING_SERVER',
        ]
        if any(x for x in luci_vars if x not in env):
            return None

        project, bucket, builder = env['BUILDBUCKET_NAME'].split(':')

        bbid: int = 0
        pipeline: LuciPipeline | None = None
        try:
            bbid = int(env['BUILDBUCKET_ID'])
            pipeline = LuciPipeline.create(bbid, fake_pipeline_props)

        except ValueError:
            pass

        # Logic to identify cas instance from swarming server is derived from
        # https://chromium.googlesource.com/infra/luci/recipes-py/+/main/recipe_modules/cas/api.py
        swarm_server = env['SWARMING_SERVER']
        cas_project = urllib.parse.urlparse(swarm_server).netloc.split('.')[0]
        cas_instance = f'projects/{cas_project}/instances/default_instance'

        result = LuciContext(
            buildbucket_id=bbid,
            build_number=int(env['BUILD_NUMBER']),
            project=project,
            bucket=bucket,
            builder=builder,
            swarming_server=env['SWARMING_SERVER'],
            swarming_task_id=env['SWARMING_TASK_ID'],
            cas_instance=cas_instance,
            pipeline=pipeline,
            triggers=LuciTrigger.create_from_environment(env),
            context_file=Path(env['LUCI_CONTEXT']),
        )
        _LOG.debug('%r', result)
        return result

    @staticmethod
    def create_for_testing(**kwargs):
        env = {
            'BUILDBUCKET_ID': '881234567890',
            'BUILDBUCKET_NAME': 'pigweed:bucket.try:builder-name',
            'BUILD_NUMBER': '123',
            'LUCI_CONTEXT': '/path/to/context/file.json',
            'SWARMING_SERVER': 'https://chromium-swarm.appspot.com',
            'SWARMING_TASK_ID': 'cd2dac62d2',
        }
        env.update(kwargs)

        return LuciContext.create_from_environment(env, {})


@dataclasses.dataclass
class FormatContext:
    """Context passed into formatting helpers.

    This class is a subset of PresubmitContext containing only what's needed by
    formatters.

    For full documentation on the members see the PresubmitContext section of
    pw_presubmit/docs.rst.

    Attributes:
        root: Source checkout root directory
        output_dir: Output directory for this specific language.
        paths: Modified files for the presubmit step to check (often used in
            formatting steps but ignored in compile steps).
        package_root: Root directory for pw package installations.
        format_options: Formatting options, derived from pigweed.json.
        dry_run: Whether to just report issues or also fix them.
    """

    root: Path | None
    output_dir: Path
    paths: tuple[Path, ...]
    package_root: Path
    format_options: FormatOptions
    dry_run: bool = False

    def append_check_command(self, *command_args, **command_kwargs) -> None:
        """Empty append_check_command."""


class PresubmitFailure(Exception):
    """Optional exception to use for presubmit failures."""

    def __init__(
        self,
        description: str = '',
        path: Path | None = None,
        line: int | None = None,
    ):
        line_part: str = ''
        if line is not None:
            line_part = f'{line}:'
        super().__init__(
            f'{path}:{line_part} {description}' if path else description
        )


@dataclasses.dataclass
class PresubmitContext:  # pylint: disable=too-many-instance-attributes
    """Context passed into presubmit checks.

    For full documentation on the members see pw_presubmit/docs.rst.

    Attributes:
        root: Source checkout root directory.
        repos: Repositories (top-level and submodules) processed by
            `pw presubmit`.
        output_dir: Output directory for this specific presubmit step.
        failure_summary_log: Path where steps should write a brief summary of
            any failures encountered for use by other tooling.
        paths: Modified files for the presubmit step to check (often used in
            formatting steps but ignored in compile steps).
        all_paths: All files in the tree.
        package_root: Root directory for pw package installations.
        override_gn_args: Additional GN args processed by `build.gn_gen()`.
        luci: Information about the LUCI build or None if not running in LUCI.
        format_options: Formatting options, derived from pigweed.json.
        num_jobs: Number of jobs to run in parallel.
        continue_after_build_error: For steps that compile, don't exit on the
            first compilation error.
        rng_seed: Seed for a random number generator, for the few steps that
            need one.
        full: Whether this is a full or incremental presubmit run.
        _failed: Whether the presubmit step in question has failed. Set to True
            by calling ctx.fail().
        dry_run: Whether to actually execute commands or just log them.
        use_remote_cache: Whether to tell the build system to use RBE.
        pw_root: The path to the Pigweed repository.
    """

    root: Path
    repos: tuple[Path, ...]
    output_dir: Path
    failure_summary_log: Path
    paths: tuple[Path, ...]
    all_paths: tuple[Path, ...]
    package_root: Path
    luci: LuciContext | None
    override_gn_args: dict[str, str]
    format_options: FormatOptions
    num_jobs: int | None = None
    continue_after_build_error: bool = False
    rng_seed: int = 1
    full: bool = False
    _failed: bool = False
    dry_run: bool = False
    use_remote_cache: bool = False
    pw_root: Path = pw_cli.env.pigweed_environment().PW_ROOT

    @property
    def failed(self) -> bool:
        return self._failed

    @property
    def incremental(self) -> bool:
        return not self.full

    def fail(
        self,
        description: str,
        path: Path | None = None,
        line: int | None = None,
    ):
        """Add a failure to this presubmit step.

        If this is called at least once the step fails, but not immediately—the
        check is free to continue and possibly call this method again.
        """
        _LOG.warning('%s', PresubmitFailure(description, path, line))
        self._failed = True

    @staticmethod
    def create_for_testing(**kwargs):
        parsed_env = pw_cli.env.pigweed_environment()
        root = parsed_env.PW_PROJECT_ROOT
        presubmit_root = root / 'out' / 'presubmit'
        presubmit_kwargs = {
            'root': root,
            'repos': (root,),
            'output_dir': presubmit_root / 'test',
            'failure_summary_log': presubmit_root / 'failure-summary.log',
            'paths': (root / 'foo.cc', root / 'foo.py'),
            'all_paths': (root / 'BUILD.gn', root / 'foo.cc', root / 'foo.py'),
            'package_root': root / 'environment' / 'packages',
            'luci': None,
            'override_gn_args': {},
            'format_options': FormatOptions(),
        }
        presubmit_kwargs.update(kwargs)
        return PresubmitContext(**presubmit_kwargs)

    def append_check_command(
        self,
        *command_args,
        call_annotation: dict[Any, Any] | None = None,
        **command_kwargs,
    ) -> None:
        """Save a subprocess command annotation to this presubmit context.

        This is used to capture commands that will be run for display in ``pw
        presubmit --dry-run.``

        Args:

            command_args: All args that would normally be passed to
                subprocess.run

            call_annotation: Optional key value pairs of data to save for this
                command. Examples:

                ::

                   call_annotation={'pw_package_install': 'teensy'}
                   call_annotation={'build_system': 'bazel'}
                   call_annotation={'build_system': 'ninja'}

            command_kwargs: keyword args that would normally be passed to
                subprocess.run.
        """
        call_annotation = call_annotation if call_annotation else {}
        calling_func: str | None = None
        calling_check = None

        # Loop through the current call stack looking for `self`, and stopping
        # when self is a Check() instance and if the __call__ or _try_call
        # functions are in the stack.

        # This used to be an isinstance(obj, Check) call, but it was changed to
        # this so Check wouldn't need to be imported here. Doing so would create
        # a dependency loop.
        def is_check_object(obj):
            return getattr(obj, '_is_presubmit_check_object', False)

        for frame_info in inspect.getouterframes(inspect.currentframe()):
            self_obj = frame_info.frame.f_locals.get('self', None)
            if (
                self_obj
                and is_check_object(self_obj)
                and frame_info.function in ['_try_call', '__call__']
            ):
                calling_func = frame_info.function
                calling_check = self_obj

        save_check_trace(
            self.output_dir,
            PresubmitCheckTrace(
                self,
                calling_check,
                calling_func,
                command_args,
                command_kwargs,
                call_annotation,
            ),
        )

    def __post_init__(self) -> None:
        PRESUBMIT_CONTEXT.set(self)

    def __hash__(self):
        return hash(
            tuple(
                tuple(attribute.items())
                if isinstance(attribute, dict)
                else attribute
                for attribute in dataclasses.astuple(self)
            )
        )


PRESUBMIT_CONTEXT: ContextVar[PresubmitContext | None] = ContextVar(
    'pw_presubmit_context', default=None
)


def get_presubmit_context():
    return PRESUBMIT_CONTEXT.get()


class PresubmitCheckTraceType(enum.Enum):
    BAZEL = 'BAZEL'
    CMAKE = 'CMAKE'
    GN_NINJA = 'GN_NINJA'
    PW_PACKAGE = 'PW_PACKAGE'


class PresubmitCheckTrace(NamedTuple):
    ctx: PresubmitContext
    check: Check | None
    func: str | None
    args: Iterable[Any]
    kwargs: dict[Any, Any]
    call_annotation: dict[Any, Any]

    def __repr__(self) -> str:
        return f'''CheckTrace(
  ctx={self.ctx.output_dir}
  id(ctx)={id(self.ctx)}
  check={self.check}
  args={self.args}
  kwargs={self.kwargs.keys()}
  call_annotation={self.call_annotation}
)'''


def save_check_trace(output_dir: Path, trace: PresubmitCheckTrace) -> None:
    trace_key = str(output_dir.resolve())
    trace_list = PRESUBMIT_CHECK_TRACE.get().get(trace_key, [])
    trace_list.append(trace)
    PRESUBMIT_CHECK_TRACE.get()[trace_key] = trace_list


def get_check_traces(ctx: PresubmitContext) -> list[PresubmitCheckTrace]:
    trace_key = str(ctx.output_dir.resolve())
    return PRESUBMIT_CHECK_TRACE.get().get(trace_key, [])


def log_check_traces(ctx: PresubmitContext) -> None:
    traces = PRESUBMIT_CHECK_TRACE.get()

    for _output_dir, check_traces in traces.items():
        for check_trace in check_traces:
            if check_trace.ctx != ctx:
                continue

            quoted_command_args = ' '.join(
                shlex.quote(str(arg)) for arg in check_trace.args
            )
            _LOG.info(
                '%s %s',
                _COLOR.blue('Run ==>'),
                quoted_command_args,
            )


def apply_exclusions(
    ctx: PresubmitContext,
    paths: Sequence[Path] | None = None,
) -> tuple[Path, ...]:
    return ctx.format_options.filter_paths(paths or ctx.paths)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of the Raspberry Pi Pico SDK."""

from contextlib import contextmanager
import logging
import os
from pathlib import Path
from typing import Sequence
import subprocess

import pw_package.git_repo
import pw_package.package_manager

_LOG = logging.getLogger(__package__)


@contextmanager
def change_working_dir(directory: Path):
    original_dir = Path.cwd()
    try:
        os.chdir(directory)
        yield directory
    finally:
        os.chdir(original_dir)


class PiPicoSdk(pw_package.package_manager.Package):
    """Install and check status of the Raspberry Pi Pico SDK."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='pico_sdk', **kwargs)
        self._pico_sdk = pw_package.git_repo.GitRepo(
            name='pico_sdk',
            url=(
                'https://pigweed.googlesource.com/'
                'third_party/github/raspberrypi/pico-sdk'
            ),
            commit='6a7db34ff63345a7badec79ebea3aaef1712f374',
        )

    def install(self, path: Path) -> None:
        self._pico_sdk.install(path)

        # Run submodule update --init to fetch tinyusb.
        with change_working_dir(path) as _pico_sdk_repo:
            command = ['git', 'submodule', 'update', '--init']
            _LOG.info('==> %s', ' '.join(command))
            subprocess.run(command)

    def info(self, path: Path) -> Sequence[str]:
        return (
            f'{self.name} installed in: {path}',
            "Enable by running 'gn args out' and adding this line:",
            f'  PICO_SRC_DIR = "{path}"',
        )


pw_package.package_manager.register(PiPicoSdk)



#!/usr/bin/env python3
# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Facilitates automating unit tests on devices with serial ports.

This library assumes that the on-device test runner emits the test results
as plain-text over a serial port, and that tests are triggered by a pre-defined
input (``DEFAULT_TEST_START_CHARACTER``) over the same serial port that results
are emitted from.
"""

import abc
import logging
from pathlib import Path

import serial  # type: ignore


_LOG = logging.getLogger("serial_test_runner")

# Verification of test pass/failure depends on these strings. If the formatting
# or output of the simple_printing_event_handler changes, this may need to be
# updated.
_TESTS_STARTING_STRING = b'[==========] Running all tests.'
_TESTS_DONE_STRING = b'[==========] Done running all tests.'
_TEST_FAILURE_STRING = b'[  FAILED  ]'

# Character used to trigger test start.
DEFAULT_TEST_START_CHARACTER = ' '.encode('utf-8')


class FlashingFailure(Exception):
    """A simple exception to be raised when flashing fails."""


class TestingFailure(Exception):
    """A simple exception to be raised when a testing step fails."""


class DeviceNotFound(Exception):
    """A simple exception to be raised when unable to connect to a device."""


class SerialTestingDevice(abc.ABC):
    """A device that supports automated testing via parsing serial output."""

    @abc.abstractmethod
    def load_binary(self, binary: Path) -> bool:
        """Flashes the specified binary to this device.

        Raises:
          DeviceNotFound: This device is no longer available.
          FlashingFailure: The binary could not be flashed.

        Returns:
          True if the binary was loaded successfully.
        """

    @abc.abstractmethod
    def serial_port(self) -> str:
        """Returns the name of the com port this device is enumerated on.

        Raises:
          DeviceNotFound: This device is no longer available.
        """

    @abc.abstractmethod
    def baud_rate(self) -> int:
        """Returns the baud rate to use when connecting to this device.

        Raises:
          DeviceNotFound: This device is no longer available.
        """


def _log_subprocess_output(level, output: bytes, logger: logging.Logger):
    """Logs subprocess output line-by-line."""

    lines = output.decode('utf-8', errors='replace').splitlines()
    for line in lines:
        logger.log(level, line)


def trigger_test_run(
    port: str,
    baud_rate: int,
    test_timeout: float,
    trigger_data: bytes = DEFAULT_TEST_START_CHARACTER,
) -> bytes:
    """Triggers a test run, and returns captured test results."""

    serial_data = bytearray()
    device = serial.Serial(baudrate=baud_rate, port=port, timeout=test_timeout)
    if not device.is_open:
        raise TestingFailure('Failed to open device')

    # Flush input buffer and trigger the test start.
    device.reset_input_buffer()
    device.write(trigger_data)

    # Block and wait for the first byte.
    serial_data += device.read()
    if not serial_data:
        raise TestingFailure('Device not producing output')

    # Read with a reasonable timeout until we stop getting characters.
    while True:
        bytes_read = device.readline()
        if not bytes_read:
            break
        serial_data += bytes_read
        if serial_data.rfind(_TESTS_DONE_STRING) != -1:
            # Set to much more aggressive timeout since the last one or two
            # lines should print out immediately. (one line if all fails or all
            # passes, two lines if mixed.)
            device.timeout = 0.01

    # Remove carriage returns.
    serial_data = serial_data.replace(b"\r", b"")

    # Try to trim captured results to only contain most recent test run.
    test_start_index = serial_data.rfind(_TESTS_STARTING_STRING)
    return (
        serial_data
        if test_start_index == -1
        else serial_data[test_start_index:]
    )


def handle_test_results(
    test_output: bytes, logger: logging.Logger = _LOG
) -> None:
    """Parses test output to determine whether tests passed or failed.

    Raises:
      TestingFailure if any tests fail or if test results are incomplete.
    """

    if test_output.find(_TESTS_STARTING_STRING) == -1:
        raise TestingFailure('Failed to find test start')

    if test_output.rfind(_TESTS_DONE_STRING) == -1:
        _log_subprocess_output(logging.INFO, test_output, logger)
        raise TestingFailure('Tests did not complete')

    if test_output.rfind(_TEST_FAILURE_STRING) != -1:
        _log_subprocess_output(logging.INFO, test_output, logger)
        raise TestingFailure('Test suite had one or more failures')

    _log_subprocess_output(logging.DEBUG, test_output, logger)

    logger.info('Test passed!')


def run_device_test(
    device: SerialTestingDevice,
    binary: Path,
    test_timeout: float,
    logger: logging.Logger = _LOG,
) -> bool:
    """Runs tests on a device.

    When a unit test run fails, results are logged as an error.

    Args:
      device: The device to run tests on.
      binary: The binary containing tests to flash on the device.
      test_timeout: If the device stops producing output longer than this
        timeout, the test is considered stuck and is aborted.

    Returns:
      True if all tests passed.
    """

    logger.info('Flashing binary to device')
    device.load_binary(binary)
    try:
        logger.info('Running test')
        test_output = trigger_test_run(
            device.serial_port(), device.baud_rate(), test_timeout
        )
        if test_output:
            handle_test_results(test_output, logger)
    except TestingFailure as err:
        logger.error(err)
        return False

    return True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_metric_proto/metric_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$pw_metric_proto/metric_service.proto\x12\x0fpw.metric.proto\"`\n\x06Metric\x12\x12\n\ntoken_path\x18\x01 \x03(\x07\x12\x13\n\x0bstring_path\x18\x02 \x03(\t\x12\x12\n\x08\x61s_float\x18\x03 \x01(\x02H\x00\x12\x10\n\x06\x61s_int\x18\x04 \x01(\rH\x00\x42\x07\n\x05value\"9\n\rMetricRequest\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric\":\n\x0eMetricResponse\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric2[\n\rMetricService\x12J\n\x03Get\x12\x1e.pw.metric.proto.MetricRequest\x1a\x1f.pw.metric.proto.MetricResponse\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_metric_proto.metric_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_METRIC']._serialized_start=57
  _globals['_METRIC']._serialized_end=153
  _globals['_METRICREQUEST']._serialized_start=155
  _globals['_METRICREQUEST']._serialized_end=212
  _globals['_METRICRESPONSE']._serialized_start=214
  _globals['_METRICRESPONSE']._serialized_end=272
  _globals['_METRICSERVICE']._serialized_start=274
  _globals['_METRICSERVICE']._serialized_end=365
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/tuf.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cpw_software_update/tuf.proto\x12\x12pw.software_update\x1a\x1fgoogle/protobuf/timestamp.proto\"i\n\x12SignedRootMetadata\x12 \n\x18serialized_root_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"s\n\x17SignedTimestampMetadata\x12%\n\x1dserialized_timestamp_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"q\n\x16SignedSnapshotMetadata\x12$\n\x1cserialized_snapshot_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"o\n\x15SignedTargetsMetadata\x12#\n\x1bserialized_targets_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"\x91\x01\n\x0e\x43ommonMetadata\x12\x14\n\x0cspec_version\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x30\n\x07\x65xpires\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x12\x11\n\x04role\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_expiresB\x07\n\x05_role\"\xed\x03\n\x0cRootMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x1b\n\x13\x63onsistent_snapshot\x18\x02 \x01(\x08\x12,\n\x04keys\x18\x03 \x03(\x0b\x32\x1e.pw.software_update.KeyMapping\x12L\n\x1aroot_signature_requirement\x18\x04 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12Q\n\x1ftimestamp_signature_requirement\x18\x05 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12P\n\x1esnapshot_signature_requirement\x18\x06 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12O\n\x1dtargets_signature_requirement\x18\x07 \x01(\x0b\x32(.pw.software_update.SignatureRequirementJ\x04\x08\x08\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\xa0\x01\n\x11TimestampMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12;\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x9e\x01\n\x10SnapshotMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12:\n\x10targets_metadata\x18\x02 \x03(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x97\x01\n\x0fTargetsMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x34\n\x0ctarget_files\x18\x02 \x03(\x0b\x32\x1e.pw.software_update.TargetFileJ\x04\x08\t\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"(\n\tSignature\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12\x0b\n\x03sig\x18\x02 \x01(\x0c\"B\n\nKeyMapping\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12$\n\x03key\x18\x02 \x01(\x0b\x32\x17.pw.software_update.Key\"s\n\x03Key\x12-\n\x08key_type\x18\x01 \x01(\x0e\x32\x1b.pw.software_update.KeyType\x12-\n\x06scheme\x18\x02 \x01(\x0e\x32\x1d.pw.software_update.KeyScheme\x12\x0e\n\x06keyval\x18\x03 \x01(\x0c\":\n\x14SignatureRequirement\x12\x0f\n\x07key_ids\x18\x01 \x03(\x0c\x12\x11\n\tthreshold\x18\x02 \x01(\r\"H\n\x04Hash\x12\x32\n\x08\x66unction\x18\x01 \x01(\x0e\x32 .pw.software_update.HashFunction\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"l\n\nTargetFile\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x0e\n\x06length\x18\x02 \x01(\x04\x12(\n\x06hashes\x18\x03 \x03(\x0b\x32\x18.pw.software_update.HashJ\x04\x08\x04\x10\x10J\x04\x08\x10\x10 J\x05\x08 \x10\x80\x02\"\x8f\x01\n\x0cMetadataFile\x12\x16\n\tfile_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x13\n\x06length\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12(\n\x06hashes\x18\x04 \x03(\x0b\x32\x18.pw.software_update.HashB\x0c\n\n_file_nameB\t\n\x07_length*5\n\x0cHashFunction\x12\x19\n\x15UNKNOWN_HASH_FUNCTION\x10\x00\x12\n\n\x06SHA256\x10\x01*N\n\x07KeyType\x12\x14\n\x10UNKNOWN_KEY_TYPE\x10\x00\x12\x07\n\x03RSA\x10\x01\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x02\x12\x17\n\x13\x45\x43\x44SA_SHA2_NISTP256\x10\x03*u\n\tKeyScheme\x12\x16\n\x12UNKNOWN_KEY_SCHEME\x10\x00\x12\x1c\n\x18RSASSA_PSS_SHA256_SCHEME\x10\x01\x12\x12\n\x0e\x45\x44\x32\x35\x35\x31\x39_SCHEME\x10\x02\x12\x1e\n\x1a\x45\x43\x44SA_SHA2_NISTP256_SCHEME\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.tuf_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_HASHFUNCTION']._serialized_start=2276
  _globals['_HASHFUNCTION']._serialized_end=2329
  _globals['_KEYTYPE']._serialized_start=2331
  _globals['_KEYTYPE']._serialized_end=2409
  _globals['_KEYSCHEME']._serialized_start=2411
  _globals['_KEYSCHEME']._serialized_end=2528
  _globals['_SIGNEDROOTMETADATA']._serialized_start=85
  _globals['_SIGNEDROOTMETADATA']._serialized_end=190
  _globals['_SIGNEDTIMESTAMPMETADATA']._serialized_start=192
  _globals['_SIGNEDTIMESTAMPMETADATA']._serialized_end=307
  _globals['_SIGNEDSNAPSHOTMETADATA']._serialized_start=309
  _globals['_SIGNEDSNAPSHOTMETADATA']._serialized_end=422
  _globals['_SIGNEDTARGETSMETADATA']._serialized_start=424
  _globals['_SIGNEDTARGETSMETADATA']._serialized_end=535
  _globals['_COMMONMETADATA']._serialized_start=538
  _globals['_COMMONMETADATA']._serialized_end=683
  _globals['_ROOTMETADATA']._serialized_start=686
  _globals['_ROOTMETADATA']._serialized_end=1179
  _globals['_TIMESTAMPMETADATA']._serialized_start=1182
  _globals['_TIMESTAMPMETADATA']._serialized_end=1342
  _globals['_SNAPSHOTMETADATA']._serialized_start=1345
  _globals['_SNAPSHOTMETADATA']._serialized_end=1503
  _globals['_TARGETSMETADATA']._serialized_start=1506
  _globals['_TARGETSMETADATA']._serialized_end=1657
  _globals['_SIGNATURE']._serialized_start=1659
  _globals['_SIGNATURE']._serialized_end=1699
  _globals['_KEYMAPPING']._serialized_start=1701
  _globals['_KEYMAPPING']._serialized_end=1767
  _globals['_KEY']._serialized_start=1769
  _globals['_KEY']._serialized_end=1884
  _globals['_SIGNATUREREQUIREMENT']._serialized_start=1886
  _globals['_SIGNATUREREQUIREMENT']._serialized_end=1944
  _globals['_HASH']._serialized_start=1946
  _globals['_HASH']._serialized_end=2018
  _globals['_TARGETFILE']._serialized_start=2020
  _globals['_TARGETFILE']._serialized_end=2128
  _globals['_METADATAFILE']._serialized_start=2131
  _globals['_METADATAFILE']._serialized_end=2274
# @@protoc_insertion_point(module_scope)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Helpful commands for working with a Git repository."""

from pathlib import Path
import subprocess
from typing import Collection, Pattern

from pw_cli import git_repo
from pw_cli.decorators import deprecated
from pw_presubmit.tools import PresubmitToolRunner

# Moved to pw_cli.git_repo.
TRACKING_BRANCH_ALIAS = git_repo.TRACKING_BRANCH_ALIAS


class LoggingGitRepo(git_repo.GitRepo):
    """A version of GitRepo that defaults to using a PresubmitToolRunner."""

    def __init__(self, repo_root: Path):
        super().__init__(repo_root, PresubmitToolRunner())


@deprecated('Extend GitRepo to expose needed functionality.')
def git_stdout(
    *args: Path | str, show_stderr=False, repo: Path | str = '.'
) -> str:
    """Runs a git command in the specified repo."""
    tool_runner = PresubmitToolRunner()
    return (
        tool_runner(
            'git',
            ['-C', str(repo), *args],
            stderr=None if show_stderr else subprocess.DEVNULL,
            check=True,
            pw_presubmit_ignore_dry_run=True,
        )
        .stdout.decode()
        .strip()
    )


def find_git_repo(path_in_repo: Path) -> git_repo.GitRepo:
    """Tries to find the root of the git repo that owns path_in_repo."""
    return git_repo.find_git_repo(path_in_repo, PresubmitToolRunner())


@deprecated('Use GitRepo().tracking_branch().')
def tracking_branch(
    repo_path: Path | None = None,
    fallback: str | None = None,
) -> str | None:
    """Returns the tracking branch of the current branch.

    Since most callers of this function can safely handle a return value of
    None, suppress exceptions and return None if there is no tracking branch.

    Args:
      repo_path: repo path from which to run commands; defaults to Path.cwd()

    Raises:
      ValueError: if repo_path is not in a Git repository

    Returns:
      the remote tracking branch name or None if there is none
    """
    repo = repo_path if repo_path is not None else Path.cwd()
    return find_git_repo(repo).tracking_branch(fallback)


@deprecated('Use GitRepo().list_files().')
def list_files(
    commit: str | None = None,
    pathspecs: Collection[Path | str] = (),
    repo_path: Path | None = None,
) -> list[Path]:
    """Lists files with git ls-files or git diff --name-only.

    Args:
      commit: commit to use as a base for git diff
      pathspecs: Git pathspecs to use in git ls-files or diff
      repo_path: repo path from which to run commands; defaults to Path.cwd()

    Returns:
      A sorted list of absolute paths
    """
    repo = repo_path if repo_path is not None else Path.cwd()
    return find_git_repo(repo).list_files(commit, pathspecs)


@deprecated('Use GitRepo.has_uncommitted_changes().')
def has_uncommitted_changes(repo: Path | None = None) -> bool:
    """Returns True if the Git repo has uncommitted changes in it.

    This does not check for untracked files.
    """
    return LoggingGitRepo(
        repo if repo is not None else Path.cwd()
    ).has_uncommitted_changes()


def describe_files(
    git_root: Path,  # pylint: disable=unused-argument
    working_dir: Path,
    commit: str | None,
    pathspecs: Collection[Path | str],
    exclude: Collection[Pattern],
    project_root: Path | None = None,
) -> str:
    """Completes 'Doing something to ...' for a set of files in a Git repo."""
    return git_repo.describe_git_pattern(
        working_dir=working_dir,
        commit=commit,
        pathspecs=pathspecs,
        exclude=exclude,
        tool_runner=PresubmitToolRunner(),
        project_root=project_root,
    )


@deprecated('Use find_git_repo().root().')
def root(repo_path: Path | str = '.') -> Path:
    """Returns the repository root as an absolute path.

    Raises:
      FileNotFoundError: the path does not exist
      subprocess.CalledProcessError: the path is not in a Git repo
    """
    return find_git_repo(Path(repo_path)).root()


def within_repo(repo_path: Path | str = '.') -> Path | None:
    """Similar to root(repo_path), returns None if the path is not in a repo."""
    try:
        return root(repo_path)
    except git_repo.GitError:
        return None


def is_repo(repo_path: Path | str = '.') -> bool:
    """True if the path is tracked by a Git repo."""
    return within_repo(repo_path) is not None


def path(
    repo_path: Path | str,
    *additional_repo_paths: Path | str,
    repo: Path | str = '.',
) -> Path:
    """Returns a path relative to a Git repository's root."""
    return root(repo).joinpath(repo_path, *additional_repo_paths)


@deprecated('Use GitRepo.commit_message().')
def commit_message(commit: str = 'HEAD', repo: Path | str = '.') -> str:
    """Returns the commit message of the revision."""
    return LoggingGitRepo(Path(repo)).commit_message(commit)


@deprecated('Use GitRepo.commit_author().')
def commit_author(commit: str = 'HEAD', repo: Path | str = '.') -> str:
    """Returns the commit author of the revision."""
    return LoggingGitRepo(Path(repo)).commit_author(commit)


@deprecated('Use GitRepo.commit_hash().')
def commit_hash(
    rev: str = 'HEAD', short: bool = True, repo: Path | str = '.'
) -> str:
    """Returns the commit hash of the revision."""
    return LoggingGitRepo(Path(repo)).commit_hash(rev, short)


@deprecated('Use GitRepo.list_submodules().')
def discover_submodules(
    superproject_dir: Path, excluded_paths: Collection[Pattern | str] = ()
) -> list[Path]:
    """Query git and return a list of submodules in the current project.

    Args:
        superproject_dir: Path object to directory under which we are looking
                          for submodules. This will also be included in list
                          returned unless excluded.
        excluded_paths: Pattern or string that match submodules that should not
                        be returned. All matches are done on posix style paths.

    Returns:
        List of "Path"s which were found but not excluded, this includes
        superproject_dir unless excluded.
    """
    return LoggingGitRepo(superproject_dir).list_submodules(excluded_paths) + [
        superproject_dir.resolve()
    ]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/echo.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11pw_rpc/echo.proto\x12\x06pw.rpc\"\x1a\n\x0b\x45\x63hoMessage\x12\x0b\n\x03msg\x18\x01 \x01(\t2A\n\x0b\x45\x63hoService\x12\x32\n\x04\x45\x63ho\x12\x13.pw.rpc.EchoMessage\x1a\x13.pw.rpc.EchoMessage\"\x00\x42\x1a\n\x18\x64\x65v.pigweed.pw_rpc.protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_rpc.echo_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030dev.pigweed.pw_rpc.proto'
  _globals['_ECHOMESSAGE']._serialized_start=29
  _globals['_ECHOMESSAGE']._serialized_end=55
  _globals['_ECHOSERVICE']._serialized_start=57
  _globals['_ECHOSERVICE']._serialized_end=122
# @@protoc_insertion_point(module_scope)
